#include <time.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "DPF_common.h"
#include "DPF_client.h"

extern DPF_fdInfo _DPF_fd[];
extern char *_client_status_str[];

// State-transition machine for TCP_SERVER
void
TCP_SERVER_state_machine(int fd, time_t current)
{
    DPF_TCP_sock *sock = (DPF_TCP_sock *) _DPF_fd[fd].info;

    int newfd;
    struct sockaddr_in peer;
    socklen_t addrlen = sizeof(struct sockaddr_in);
    DPF_TCP_sock *new_sock;

    while(true) {
        LOG(D_FULL, "[TCP_SERVER, fd=%d] %s\n", fd, _client_status_str[sock->status]);
        switch(sock->status) {
            case LOCAL_LISTEN:
            case FULL_LISTEN:
            case LISTENING_1:
            case LISTENING_2:
                // Condition to enter here:
                //    - sock->fd is read ready
                // Action: accept a new connection and queue it
                // 1. accept the connection
                newfd = accept(fd, (struct sockaddr *)&peer, &addrlen);
                if (newfd < 0) {
                    if (errno == EAGAIN) return;
                    sock->error = errno;
                    return;
                }
                LOG(D_FULL, "%d = accept(%d, %s)\n", newfd, fd, _sin2str(&peer));
                // 2. make newfd nonblock
                if (_make_nonblock(newfd) < 0) {
                    LOG(D_BASIC, "failed to make newfd=%d nonblocking\n", newfd);
                    close(newfd);
                    return;
                }
                // 3. register HIDDEN sock
                new_sock = new DPF_TCP_sock(newfd);
                new_sock->listener = fd;
                new_sock->peer = new ExAddr(&peer, NULL, NULL);
                new_sock->status = CONNECTION_SUCC;
                if (!(new_sock->peer)) { ABORT("new DPF_peer failed"); }
                _register_sock(newfd, HIDDEN, new_sock);
                // 4. queue the connection
                new_sock->status = CONNECTION_SUCC;
                _queue_connection(sock, new_sock);
                return;

            default:
                ABORT("TCP_SERVER_state_machine got to unknown state\n");
        }
    } // while(true)
}
