#include <stdio.h>
#include <string.h>
#include <time.h>

#include "DPF_client.h"

DPF_TCP_sock::DPF_TCP_sock(int owner)
    : ClientSock(owner)
{
    listener = -1;
    peer = NULL;
    chead = ctail = NULL;
}

DPF_TCP_sock::~DPF_TCP_sock()
{
    if (peer) { delete peer; }
    // Since the connection queue is a list of DPF_TCP_sock objects, the queue must
    // be deleted by some external functions: delete_connection_queue @ close.C
}

int
DPF_TCP_sock::make_fresh()
{
    if (myaddr) {
        delete myaddr;
        myaddr = NULL;
    }
    if (inagent) {
        _close_duped_fd(inagent->get_fd());
        delete inagent;
        inagent = NULL;
    }
    if (peer_agent) {
        _close_duped_fd(peer_agent->get_fd());
        delete peer_agent;
        peer_agent = NULL;
    }

    int tmp_sock = _tcp_nonblock_socket();
    if (tmp_sock < 0) {
        LOG(D_BASIC, "make_fresh:temporary socket creation failed\n");
        return -1;
    }
    if (_replace_fd(fd, tmp_sock) < 0) {
        LOG(D_BASIC, "can't replace DPF_TCP_sock->fd with temp sock\n");
        return -1;
    }
    return 0;
}

void
DPF_TCP_sock::tostr(const char *prefix, char *str)
{
    ClientSock::tostr(prefix, str);
    sprintf(&str[strlen(str)], "\n%speer=", prefix);
    if (peer) {
        peer->tostr("", &str[strlen(str)]);
    } else {
        strcat(str, "NULL");
    }
    sprintf(&str[strlen(str)], " listener=%d\n", listener);
    strcat(str, prefix);
    strcat(str, "Conn_Q: ");
    DPF_TCP_sock *ptr;
    for (ptr = chead; ptr; ptr = ptr->chead) {
        sprintf(&str[strlen(str)], "<%d, ", ptr->fd);
        ptr->peer->tostr("", &str[strlen(str)]);
        strcat(str, "> ");
    }
}

int
DPF_TCP_sock::stripe(char *buf, bool deep)
{
    // length of the serialization excluding this field(2 bytes)
    // listener(4 bytes)
    // ExAddr peer
        // peer serialization length. 0 means that peer = NULL(2 bytes)
        // peer serialization(length bytes)
    // if (deep == true)
        // the number of entries in the connection queue(2 bytes)
        // the concatenation of serializations of entries in the connection queue
        // using DPF_TCP_sock::stripe(buf, deep=false)
    // the serialization of ClientSock

    char str[5000];
    uint32_t nlong;
    uint16_t nshort;
    int index = 2;
    // listener(4 bytes)
    nlong = htonl(listener);
    memcpy(&buf[index], &nlong, 4);
    sprintf(str, "[%d]listener=%d", index, listener);
    index += 4;
    // ExAddr peer
        // peer serialization length. 0 means that peer = NULL(2 bytes)
        // peer serialization(length bytes)
    sprintf(&str[strlen(str)], "[%d]peer_serialization", index);
    if (peer) {
        int peer_len = peer->stripe(&buf[index]);
        index += peer_len;
    } else {
        nshort = htons(0);
        memcpy(&buf[index], &nshort, 2);
        index += 2;
    }
    // if (deep == true)
        // the number of entries in the connection queue(2 bytes)
        // the concatenation of serializations of entries in the connection queue
        // using DPF_TCP_sock::stripe(buf, deep=false)
    if (deep) {
        int Qlen;
        int saved_idx = index;
        sprintf(&str[strlen(str)], "[%d]Conn_Q", index);
        index += 2;
        DPF_TCP_sock *conn;
        for (conn = chead, Qlen = 0; conn; conn = conn->chead, Qlen++) {
            sprintf(&str[strlen(str)], "[%d]Queue_entry", index);
            int conn_len = conn->stripe(&buf[index], false);
            index += conn_len;
        }
        nshort = htons(Qlen);
        memcpy(&buf[saved_idx], &nshort, 2);
    }
    // the serialization of ClientSock
    sprintf(&str[strlen(str)], "[%d]ClientSock_serialization", index);
    int ClientSock_len = ClientSock::stripe(&buf[index]);
    index += ClientSock_len;

    // length of the serialization excluding this field(2 bytes)
    nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of DPF_TCP_sock: [0]len=%d%s\n", index-2, str);

    return index;
}

int
DPF_TCP_sock::unstripe(char *buf, bool deep)
{
    // length of the serialization excluding this field(2 bytes)
    // listener(4 bytes)
    // ExAddr peer
        // peer serialization length. 0 means that peer = NULL(2 bytes)
        // peer serialization(length bytes)
    // if (deep == true)
        // the number of entries in the connection queue(2 bytes)
        // the concatenation of serializations of entries in the connection queue
        // using DPF_TCP_sock::stripe(buf, deep=false)
    // the serialization of ClientSock

    char str[5000];
    uint32_t nlong;
    uint16_t nshort;
    int index = 0;
    // length of the serialization excluding this field(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // listener(4 bytes)
    memcpy(&nlong, &buf[index], 4);
    listener = ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]listener=%d", index, listener);
    index += 4;
    // ExAddr peer
        // peer serialization length. 0 means that peer = NULL(2 bytes)
        // peer serialization(length bytes)
    sprintf(&str[strlen(str)], "[%d]peer_serialization", index);
    memcpy(&nshort, &buf[index], 2);
    unsigned short peer_len = ntohs(nshort);
    if (peer_len > 0) {
        if (!peer) {
            peer = new ExAddr(NULL, NULL, NULL);
        }
        index += peer->unstripe(&buf[index]);
    } else {
        index += 2;
    }
    // if (deep == true)
        // the number of entries in the connection queue(2 bytes)
        // the concatenation of serializations of entries in the connection queue
        // using DPF_TCP_sock::stripe(buf, deep=false)
    if (deep) {
        memcpy(&nshort, &buf[index], 2);
        unsigned short Qlen = ntohs(nshort);
        sprintf(&str[strlen(str)], "[%d]Conn_Q(%d entries)", index, Qlen);
        index += 2;
        for (; Qlen > 0; Qlen--) {
            // create a HIDDEN sock
            DPF_TCP_sock *hidden = new DPF_TCP_sock(-1);
            sprintf(&str[strlen(str)], "[%d]Queue_entry", index);
            int hidden_len = hidden->unstripe(&buf[index], false);
            index += hidden_len;
            // queue the HIDDEN sock
            if (chead) {
                ctail->chead = hidden; // ctail as prev pointer
                hidden->ctail = ctail; // chead as next pointer
                ctail = hidden;
            } else {
                chead = ctail = hidden;
                hidden->ctail = this;
            }
        }
    }
    // the serialization of ClientSock
    sprintf(&str[strlen(str)], "[%d]ClientSock_serialization", index);
    int ClientSock_len = ClientSock::unstripe(&buf[index]);
    index += ClientSock_len;

    LOG(D_FULL, "Unserialization of DPF_TCP_sock: %s\n", str);

    if (length != index-2) {
        ABORT("length does not match with the actual size of the serialization");
    }

    return index;
}
