#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

static int
get_connection(DPF_TCP_sock *listener, struct sockaddr *from, socklen_t *addrlen)
{
    int newfd = listener->chead->fd;
    DPF_TCP_sock *accepted = (DPF_TCP_sock *) _DPF_fd[newfd].info;
    accepted->app_status = CONNECTED;

    if (from) {
        *addrlen = sizeof(struct sockaddr_in);
        memcpy(from, accepted->peer->local, *addrlen);
    }

    // change HIDDEN to ACCEPTED
    _DPF_fd[newfd].type = ACCEPTED;

    // dequeue the connection
    if (accepted->chead) {
        listener->chead = accepted->chead;
        accepted->chead->ctail = listener;
    } else {
        listener->chead = listener->ctail = NULL;
    }
    accepted->chead = accepted->ctail = NULL;

    return newfd;
}

extern "C" {
int
DPF_accept(int fd, struct sockaddr * peer, socklen_t *addrlen)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    int newfd;
    if (_DPF_fd[fd].type == NOT_DEFINED) {
        LOG(D_FULL, "accept(%d[NOT_DEFINED], ...)\n", fd);
        newfd = accept(fd, peer, addrlen);
        LOG(D_FULL, "\treturning %d\n", newfd);
        return newfd;
    } else if (_DPF_fd[fd].type != TCP_SERVER) {
        LOG(D_BASIC, "DPF_accept called for invalid fd: %d\n", fd);
        errno = EINVAL;
        return -1;
    }

    DPF_TCP_sock *sock = (DPF_TCP_sock *) _DPF_fd[fd].info;
    LOG(D_FULL, "DPF_accept(%d[DPF_SOCKET] ...)...\n", fd);

    // if pending error, return error
    if (sock->error) {
        errno = sock->error;
        return -1;
    }

    // if connection queue is not empty, return the connection from the queue
    if (sock->chead) {
        newfd = get_connection(sock, peer, addrlen);
        return newfd;
    } else if (sock->nonblock) {
        errno = EAGAIN;
        return -1;
    }

    // now we need to wait until the queue becomes not empty
    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->error) {
            errno = sock->error;
            return -1;
        }
        if (sock->chead) {
            newfd = get_connection(sock, peer, addrlen);
            return newfd;
        }
    }
}
} // extern "C"
