#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include "DPF_client.h"

extern char *_client_status_str[];
extern char *_app_status_str[];

ClientSock::ClientSock(int fd)
{
    this->fd = fd;
    hfd = -1;
    refs = 1;
    forked = false;
    status = FRESH;
    // entered: undefined
    app_status = CREATED;
    nonblock = false;
    bound = false;
    myaddr = NULL;
    error = 0;
    inagent = NULL;
    peer_agent = NULL;
}

ClientSock::~ClientSock()
{
    // fd will get closed by _close_fd or _close_duped_fd
    if (hfd >= 0) { close(hfd); }
    if (myaddr) { delete myaddr; }
    if (inagent) { delete inagent; }
    if (peer_agent) { delete peer_agent; }
}

void
ClientSock::tostr(const char *prefix, char *str)
{
    sprintf(str, "%sfd=%d, refs=%d, status=%s, entered=%lu\n", fd, refs,
            _client_status_str[status], entered); 
    sprintf(&str[strlen(str)], "%sapp_status=%s ", prefix, _client_status_str[app_status]);
    strcat(str, "myaddr=");
    if (myaddr) {
        myaddr->tostr("", &str[strlen(str)]);
    } else {
        strcat(str, "none");
    }
    strcat(str, "\n");
    strcat(str, prefix);
    if (nonblock) {
        strcat(str, "nonblock ");
    } else {
        strcat(str, "block ");
    }
    if (bound) {
        strcat(str, "bound\n");
    } else {
        strcat(str, "not-bound\n");
    }
    sprintf(&str[strlen(str)], "%sinagent=", prefix);
    if (inagent) {
        inagent->tostr("", &str[strlen(str)]);
    } else {
        strcat(str, "NULL");
    }
    strcat(str, "\n");
    sprintf(&str[strlen(str)], "%speer_agent=", prefix);
    if (peer_agent) {
        peer_agent->tostr("", &str[strlen(str)]);
    } else {
        strcat(str, "NULL");
    }
    strcat(str, "\n");
    sprintf(&str[strlen(str)], "%serror=%d", prefix, error);
}

int
ClientSock::stripe(char *buf)
{
    // the size of the serialization excluding this field(2 bytes)
    // fd(4 bytes in network order)
    // hfd(4 bytes in network order)
    // refs(2 bytes in network order)
    // status(1 byte)
    // entered(4 bytes in network order)
    // app_status(1 byte)
    // nonblock(1 byte)
    // bound(1 byte)
    // myaddr: the serialization of ExAddr. If NULL, set 2 bytes with 0
    // inagent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    // peer_agent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    // error(4 bytes in network order)
    
    char str[1000];
    uint32_t nlong;
    uint16_t nshort;
    int index = 2;
    // fd(4 bytes in network order)
    nlong = htonl(fd);
    memcpy(&buf[index], &nlong, 4);
    sprintf(str, "[%d]fd=%d", index, fd);
    index += 4;
    // hfd(4 bytes in network order)
    nlong = htonl(hfd);
    memcpy(&buf[index], &nlong, 4);
    sprintf(&str[strlen(str)], "[%d]hfd=%d", index, hfd);
    index += 4;
    // refs(2 bytes in network order)
    nshort = htons(refs);
    memcpy(&buf[index], &nshort, 2);
    sprintf(&str[strlen(str)], "[%d]refs=%d", index, refs);
    index += 2;
    // status(1 byte)
    sprintf(&str[strlen(str)], "[%d]status=%s", index, _client_status_str[status]);
    buf[index++] = status;
    // entered(4 bytes in network order)
    nlong = htonl(entered);
    memcpy(&buf[index], &nlong, 4);
    sprintf(str, "[%d]entered=%lu", index, entered);
    index += 4;
    // app_status(1 byte)
    sprintf(&str[strlen(str)], "[%d]app_status=%s", index, _app_status_str[app_status]);
    buf[index++] = app_status;
    // nonblock(1 byte)
    sprintf(&str[strlen(str)], "[%d]nonblock=%d", index, nonblock);
    buf[index++] = (nonblock) ? 1 : 0;
    // bound(1 byte)
    sprintf(&str[strlen(str)], "[%d]bound=%d", index, bound);
    buf[index++] = (bound) ? 1 : 0;
    // myaddr: the serialization of ExAddr. If NULL, set 2 bytes with 0
    sprintf(&str[strlen(str)], "[%d]myaddr_serialization", index);
    if (myaddr) {
        int myaddr_len = myaddr->stripe(&buf[index]);
        index += myaddr_len;
    } else {
        nshort = htons(0);
        memcpy(&buf[index], &nshort, 2);
        index += 2;
    }
    // inagent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    sprintf(&str[strlen(str)], "[%d]inagent_serialization", index);
    if (inagent) {
        int inagent_len = inagent->stripe(&buf[index]);
        index += inagent_len;
    } else {
        nshort = htons(0);
        memcpy(&buf[index], &nshort, 2);
        index += 2;
    }
    // peer_agent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    sprintf(&str[strlen(str)], "[%d]peer_agent_serialization", index);
    if (peer_agent) {
        int peer_agent_len = peer_agent->stripe(&buf[index]);
        index += peer_agent_len;
    } else {
        nshort = htons(0);
        memcpy(&buf[index], &nshort, 2);
        index += 2;
    }
    // error(4 bytes)
    nlong = htonl(error);
    memcpy(&buf[index], &nlong, 4);
    sprintf(&str[strlen(str)], "[%d]error=%d", index, error);
    index += 4;

    // the size of the serialization excluding this field(2 bytes)
    nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of ClientSock: [0]len=%d%s\n", index-2, str);

    return index;
}

int
ClientSock::unstripe(char *buf)
{
    // the size of the serialization excluding this field(2 bytes)
    // fd(4 bytes in network order)
    // hfd(4 bytes in network order)
    // forked = true
    // refs(2 bytes in network order)
    // status(1 byte)
    // entered(4 bytes in network order)
    // app_status(1 byte)
    // nonblock(1 byte)
    // bound(1 byte)
    // myaddr: the serialization of ExAddr. If NULL, set 2 bytes with 0
    // inagent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    // peer_agent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    // error(4 bytes in network order)

    char str[1000];
    uint32_t nlong;
    uint16_t nshort;
    int index = 0;
    // the size of the serialization excluding this field(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // fd(4 bytes in network order)
    memcpy(&nlong, &buf[index], 4);
    fd = (int)ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]fd=%d", index, fd);
    index += 4;
    // hfd(4 bytes in network order)
    memcpy(&nlong, &buf[index], 4);
    hfd = (int)ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]hfd=%d", index, fd);
    index += 4;
    // forked = true;
    forked = true;
    // refs(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    refs = (int)ntohs(nshort);
    sprintf(&str[strlen(str)], "[%d]refs=%d", index, refs);
    index += 2;
    // status(1 byte)
    status = (DPF_client_status)buf[index];
    sprintf(&str[strlen(str)], "[%d]status=%s", index, _client_status_str[status]);
    index++;
    // entered(4 bytes in network order)
    memcpy(&nlong, &buf[index], 4);
    entered = (int)ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]entered=%d", index, entered);
    index += 4;
    // app_status(1 byte)
    app_status = (DPF_app_status)buf[index];
    sprintf(&str[strlen(str)], "[%d]app_status=%s", index, _app_status_str[app_status]);
    index++;
    // nonblock(1 byte)
    nonblock = (buf[index] == 1) ? true : false;
    sprintf(&str[strlen(str)], "[%d]nonblock=%d", index, nonblock);
    index++;
    // bound(1 byte)
    bound = (buf[index] == 1) ? true : false;
    sprintf(&str[strlen(str)], "[%d]bound=%d", index, bound);
    index++;
    // myaddr: the serialization of ExAddr. If NULL, set 2 bytes with 0
    sprintf(&str[strlen(str)], "[%d]myaddr_serialization", index);
    memcpy(&nshort, &buf[index], 2);
    unsigned short myaddr_len = ntohs(nshort);
    if (myaddr_len > 0) {
        if (!myaddr) {
            myaddr = new ExAddr(NULL, NULL, NULL);
        }
        index += myaddr->unstripe(&buf[index]);
    } else {
        index += 2;
    }
    // inagent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    sprintf(&str[strlen(str)], "[%d]inagent_serialization", index);
    memcpy(&nshort, &buf[index], 2);
    unsigned short inagent_len = ntohs(nshort);
    if (inagent_len > 0) {
        if (!inagent) {
            inagent = new DPF_channel(-1);
        }
        index += inagent->unstripe(&buf[index]);
    } else {
        index += 2;
    }
    // peer_agent: the serialization of DPF_channel. If NULL, set 2 bytes with 0
    sprintf(&str[strlen(str)], "[%d]peer_agent_serialization", index);
    memcpy(&nshort, &buf[index], 2);
    unsigned short peer_agent_len = ntohs(nshort);
    if (peer_agent_len > 0) {
        if (!peer_agent) {
            peer_agent = new DPF_channel(-1);
        }
        index += peer_agent->unstripe(&buf[index]);
    } else {
        index += 2;
    }
    // error(4 bytes)
    memcpy(&nlong, &buf[index], 4);
    error = ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]error=%d", index, error);
    index += 4;

    LOG(D_FULL, "Unserialization of ClientSock: %s\n", str);

    if (length != index-2) {
        ABORT("length does not match with the actual size of the serialization");
    }

    return index;
}
