#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];
extern int _DPF_maxfd;

static int
simple_close(int fd, ClientSock *sock)
{
    // fd adjustment
    if (sock->fd == fd) {
        for (int i=fd+1; i<_DPF_maxfd; i++) {
            if (_DPF_fd[i].type == _DPF_fd[fd].type  && _DPF_fd[i].info == sock) {
                sock->fd = i;
            }
        }
    }

    // decrease refs
    (sock->refs)--;

    return _close_duped_fd(fd);
}

static void
delete_connection_queue(DPF_TCP_sock *sock)
{
    DPF_TCP_sock *conn, *ptr;

    conn = sock->chead;
    while(conn) {
        ptr = conn;
        conn = conn->chead;
        _close_fd(ptr->fd);
    }
    sock->chead = sock->ctail = NULL;
}

static int
tcp_close(int fd)
{
    DPF_TCP_sock *sock = (DPF_TCP_sock *)_DPF_fd[fd].info;
    
    if (sock->refs > 1) {
        return simple_close(fd, sock);
    }

    if (sock->inagent) {
        _close_duped_fd(sock->inagent->get_fd());
    }
    if (sock->peer_agent) {
        _close_duped_fd(sock->peer_agent->get_fd());
    }

    if (sock->hfd >= 0) {
        if (!(sock->forked)) {
            unsigned short size;
            char buf[50];
            size = _stripe_UDP_msg(buf, TCP_OPEN_DEL, sock->myaddr->local, sock->myaddr->proxy);
            if (send(sock->hfd, buf, size, 0) < 0) {
                LOG(D_BASIC, "Can't send TCP_DEL\n");
            }
        }
        _DPF_fd[sock->hfd].type = NOT_DEFINED;
    }

    int result = 0;
    // if connection queue is not empty, return -1 
    if (sock->chead) {
        delete_connection_queue(sock);
        result = -1;
    }

    if (result < 0) {
        (void) _close_fd(fd);
    } else {
        result = _close_fd(fd);
    }

    return result;
}

static int
udp_close(int fd)
{
    DPF_UDP_sock *sock = (DPF_UDP_sock *)_DPF_fd[fd].info;

    if (sock->refs > 1) {
        return simple_close(fd, sock);
    }

    if (sock->inagent) {
        _close_duped_fd(sock->inagent->get_fd());
    }
    if (sock->peer_agent) {
        _close_duped_fd(sock->peer_agent->get_fd());
    }

    if (sock->hfd >= 0) {
        if (!(sock->forked)) {
            unsigned short size;
            char buf[50];
            size = _stripe_UDP_msg(buf, UDP_OPEN_DEL, sock->myaddr->local, sock->myaddr->proxy);
            if (send(sock->hfd, buf, size, 0) < 0) {
                LOG(D_BASIC, "Can't send TCP_DEL\n");
            }
        }
        _DPF_fd[sock->hfd].type = NOT_DEFINED;
    }

    // if rbuf is not empty, return -1
    int result;
    if (sock->rbuf.no_msg > 0) {
        result = -1;
    } else {
        result = 0;
    }

    // delete everything
    if (result < 0) {
        (void)_close_fd(fd);
    } else {
        result = _close_fd(fd);
    }

    return result;
}

extern "C" {
void
DPF_cleanup(void)
{
    for (int i = 0; i <= _DPF_maxfd; i++) {
        switch(_DPF_fd[i].type) {
            case TCP_UNCLASSIFIED:
            case TCP_SERVER:
            case TCP_CLIENT:
            case ACCEPTED:
            case UDP:
                DPF_close(i);
                break;
            default:
                break;
        }
    } // end of for
}

int
DPF_close(int fd)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // check fd
    switch(_DPF_fd[fd].type) {
        case TCP_UNCLASSIFIED:
        case TCP_SERVER:
        case TCP_CLIENT:
        case ACCEPTED:
            LOG(D_FULL, "DPF_close(%d[TCP])\n", fd);
            return tcp_close(fd);
        case UDP:
            LOG(D_FULL, "DPF_close(%d[UDP])\n", fd);
            return udp_close(fd);
        case NOT_DEFINED:
            return close(fd);

        case TCP_IN_MNGER:
        case TCP_OUT_MNGER:
        case TCP_PEER_IN_MNGER:
        case HIDDEN:
        case UDP_IN_MNGER:
        case UDP_OUT_MNGER:
        case UDP_PEER_IN_MNGER:
        case HEARTBEAT_FD:
            LOG(D_BASIC, "DPF_close(%d[MNGER | HIDDEN | HEARTBEAT])\n", fd);
            errno = EBADF;
            return -1;
        default:
            ABORT("Unknown type of socket");
    }
}
} // extern "C"
