#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {

int
DPF_dup(int fd)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

	int newfd;
    switch(_DPF_fd[fd].type) {
        case TCP_UNCLASSIFIED:
        case TCP_SERVER:
        case TCP_CLIENT:
        case UDP:
        case ACCEPTED:
            newfd = dup(fd);
            if (newfd < 0) {
                return newfd;
            }
            if (_copy_sock(fd, newfd) < 0) {
                close(newfd);
                return -1;
            }
            LOG(D_FULL, "[DPF_dup(%d, %d)]\n", fd, newfd);
            return newfd;
        case TCP_IN_MNGER:
        case TCP_OUT_MNGER:
        case TCP_PEER_IN_MNGER:
        case UDP_IN_MNGER:
        case UDP_OUT_MNGER:
        case UDP_PEER_IN_MNGER:
        case HIDDEN:
        case RESERVED:
            LOG(D_BASIC, "DPF_dup: called for invalid fd\n");
            errno = EINVAL;
            return -1;
        default:
            return dup(fd);
    }
}

int
DPF_dup2(int fd, int newfd)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    switch(_DPF_fd[fd].type) {
        case TCP_UNCLASSIFIED:
        case TCP_SERVER:
        case TCP_CLIENT:
        case UDP:
        case ACCEPTED:
            switch(_DPF_fd[newfd].type) {
                case TCP_UNCLASSIFIED:
                case TCP_SERVER:
                case TCP_CLIENT:
                case UDP:
                case ACCEPTED:
                    DPF_close(newfd);
                    if (dup2(fd, newfd) < 0) {
                        return -1;
                    }
                    if (_copy_sock(fd, newfd) < 0) {
                        close(newfd);
                        return -1;
                    }
                    LOG(D_FULL, "[DPF_dup(%d, %d)]\n", fd, newfd);
                    return newfd;
                case TCP_IN_MNGER:
                case TCP_OUT_MNGER:
                case TCP_PEER_IN_MNGER:
                case UDP_IN_MNGER:
                case UDP_OUT_MNGER:
                case UDP_PEER_IN_MNGER:
                case HIDDEN:
                case RESERVED:
                    LOG(D_BASIC, "DPF_dup: newfd is one of hidden socket\n");
                    errno = EINVAL;
                    return -1;
                default:
                    if (dup2(fd, newfd) < 0) {
                        return -1;
                    }
                    if (_copy_sock(fd, newfd) < 0) {
                        close(newfd);
                        return -1;
                    }
                    return newfd;
            }
        case TCP_IN_MNGER:
        case TCP_OUT_MNGER:
        case TCP_PEER_IN_MNGER:
        case UDP_IN_MNGER:
        case UDP_OUT_MNGER:
        case UDP_PEER_IN_MNGER:
        case HIDDEN:
        case RESERVED:
            LOG(D_BASIC, "DPF_dup: called for invalid fd\n");
            errno = EINVAL;
            return -1;
        default:
            switch(_DPF_fd[newfd].type) {
                case TCP_UNCLASSIFIED:
                case TCP_SERVER:
                case TCP_CLIENT:
                case UDP:
                case ACCEPTED:
                    LOG(D_FULL, "DPF socket(%d) get deleted due to dup2(%d, %d)\n",
                            newfd, fd, newfd);
                    DPF_close(newfd);
                    break;
                case TCP_IN_MNGER:
                case TCP_OUT_MNGER:
                case TCP_PEER_IN_MNGER:
                case UDP_IN_MNGER:
                case UDP_OUT_MNGER:
                case UDP_PEER_IN_MNGER:
                case HIDDEN:
                case RESERVED:
                    LOG(D_BASIC, "DPF_dup: newfd is one of hidden socket\n");
                    errno = EINVAL;
                    return -1;
                default:
                    break;
            }
            return dup2(fd, newfd);
    }
}

}
