#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <time.h>
#include <errno.h>

// solaris
#include <sys/types.h>
#include <sys/socket.h>
// solaris

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {

int
DPF_execv(const char *path, char *const argv[])
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    if (DPF_inheritTo() < 0) {
        LOG(D_BASIC, "inheritTo failed. sockets will not be inherited to new process\n");
    }
	return execv(path, argv);
}

int
DPF_execl(const char *path, const char *arg, ...)
{
	va_list ap;
	char *next_arg;
	char *argv[20];
	int noargs = 0;

    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    if (DPF_inheritTo() < 0) {
        LOG(D_BASIC, "inheritTo failed. sockets will not be inherited to new process\n");
    }

	argv[noargs++] = (char *)arg;

	va_start(ap, arg);
	while(1) {
		next_arg = va_arg(ap, char *);
		argv[noargs++] = next_arg;
		if (!next_arg) {
			break;
		}
	}
	va_end(ap);

	return execv(path, argv);
}

int
DPF_execve(const char *path, char *const argv [], char *const envp[])
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    if (DPF_inheritTo() < 0) {
        LOG(D_BASIC, "inheritTo failed. sockets will not be inherited to new process\n");
    }
	return execve(path, argv, envp);
}

}
