#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {

int
DPF_fcntl(int fd, int cmd, void *arg)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    int newfd;
    int val;
    ClientSock *sock;
    switch(_DPF_fd[fd].type) {
        case TCP_UNCLASSIFIED:
        case TCP_SERVER:
        case TCP_CLIENT:
        case UDP:
        case ACCEPTED:
            switch(cmd) {
                case F_DUPFD:
                    LOG(D_FULL, "[DPF_fcntl(%d, F_DUPFD, arg)]\n", fd);
                    newfd = fcntl(fd, cmd, arg);
                    if (newfd < 0) {
                        LOG(D_BASIC, "DPF_fcntl: fcntl failed: %s\n",
                                strerror(errno));
                        return -1;
                    }
                    if (_copy_sock(fd, newfd) < 0) {
                        close(newfd);
                        return -1;
                    }
                    return newfd;
                case F_SETFL:
                    sock = _DPF_fd[fd].info;
                    val = (int) arg;
                    if (val & O_NONBLOCK) {
                        LOG(D_FULL, "[DPF_fcntl(%d, F_SETFL, nonblock)]\n", fd);
                        sock->nonblock = true;
                    } else {
                        LOG(D_FULL, "[DPF_fcntl(%d, F_SETFL, block)]\n", fd);
                        sock->nonblock = false;
                        // still keep the socket nonblocking
                        val |= O_NONBLOCK;
                    }
                    return fcntl(fd, cmd, val);
                case F_GETFL:
                    sock = _DPF_fd[fd].info;
                    if ((val = fcntl(fd, cmd, 0)) < 0) {
                        LOG(D_BASIC, "DPF_fcntl: fcntl failed: %s\n",
                                strerror(errno));
                        return -1;
                    }
                    if (!(sock->nonblock)) {
                        LOG(D_FULL, "[DPF_fcntl(%d, F_GETFL, nonblock)]\n", fd);
                        val &= ~O_NONBLOCK;
                    }
                    return val;
                default:
                    return fcntl(fd, cmd, arg);
            }
        case TCP_IN_MNGER:
        case TCP_OUT_MNGER:
        case TCP_PEER_IN_MNGER:
        case UDP_IN_MNGER:
        case UDP_OUT_MNGER:
        case UDP_PEER_IN_MNGER:
        case HIDDEN:
        case RESERVED:
            LOG(D_BASIC, "DPF_fcntl: called for invalid fd\n");
            errno = EINVAL;
            return -1;
        default:
            return fcntl(fd, cmd, arg);
    }
}

} // extern "C"
