#include <string.h>
#include <sys/socket.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {

int
DPF_getsockname(int sockfd, struct sockaddr *name, socklen_t *namelen)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }
    if (_DPF_fd[sockfd].type == NOT_DEFINED) {
        return getsockname(sockfd, name, namelen);
    }

    if (*namelen < sizeof(struct sockaddr_in)) {
        LOG(D_BASIC, "DPF_getsockname: insufficient space passed\n");
        errno = EINVAL;
        return -1;
    }

    ClientSock *sock;
    DPF_TCP_sock *tcp_sock;
    switch(_DPF_fd[sockfd].type) {
        case TCP_UNCLASSIFIED:
        case TCP_SERVER:
        case TCP_CLIENT:
        case UDP:
            sock = _DPF_fd[sockfd].info;
            break;
        case ACCEPTED:
            // For accepted socket, the address of listening socket must be returned
            tcp_sock = (DPF_TCP_sock *)_DPF_fd[sockfd].info;
            sock = _DPF_fd[tcp_sock->listener].info;
            break;
        default:
            LOG(D_BASIC, "DPF_getsockname: called for invalid fd type\n");
            errno = EINVAL;
            return -1;
    }

    if (!(sock->myaddr)) {
        LOG(D_BASIC, "DPF_getsockname: sock->myaddr is NULL\n");
        errno = EINVAL;
        return -1;
    }

    if (sock->myaddr->proxy) {
        memcpy(name, sock->myaddr->proxy, sizeof(struct sockaddr_in));
        *namelen = sizeof(struct sockaddr_in);
    } else if (sock->myaddr->local) {
        memcpy(name, sock->myaddr->local, sizeof(struct sockaddr_in));
        *namelen = sizeof(struct sockaddr_in);
    } else {
        LOG(D_BASIC, "DPF_getsockname: called for socket not bound yet\n");
        bzero(name, *namelen);
    }
    return 0;
}

} // extern "C"
