#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {

int
DPF_getsockopt(
		int fd,
		int level,
		int optname,
		void *opt_val,
		socklen_t *optlen)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

	LOG(D_FULL, "[GCB_getsockopt(%d)]\n", fd);
	if (level != SOL_SOCKET || optname != SO_ERROR) {
		return getsockopt(fd, level, optname, opt_val, optlen);
    }

    int err;
    ClientSock *sock;
    switch(_DPF_fd[fd].type) {
        case TCP_CLIENT:
            // At the moment, DPF only supports the cases that the application calls
            // nonblocking connect and checks if the connection is finished by
            // calling getsockopt
            sock = _DPF_fd[fd].info;
            switch(sock->status) {
                case CONNECT_DEST_AGENT_2: // sock->peer_agent is connecting
                case CONTACT_DEST_AGENT_1: // sock->peer_agent is sending CONNECT
                case CONTACT_DEST_AGENT_2: // sock->peer_agent is receiving the reply
                case ACTIVE_CONNECTING_2: // connecting to the dest
                    err = EINPROGRESS;
                    memcpy(opt_val, &err, sizeof(int));
                    *optlen = sizeof(int);
                    return 0;
                case CONNECTION_SUCC:
                    err = 0;
                    memcpy(opt_val, &err, sizeof(int));
                    sock->app_status = CONNECTED;
                    return 0;
                default:
				    memcpy(opt_val, &sock->error, sizeof(int));
				    *optlen = sizeof(int);
                    return 0;
            }
        case NOT_DEFINED:
        case TCP_UNCLASSIFIED:
        case TCP_SERVER:
        case UDP:
        case ACCEPTED:
            sock = _DPF_fd[fd].info;
            memcpy(opt_val, &sock->error, sizeof(int));
            *optlen = sizeof(int);
            return 0;
        default:
            LOG(D_BASIC, "DPF_getsockopt: called for invalid socket\n");
            errno = EINVAL;
            return -1;
    }
}

} // extern "C"
