#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <netdb.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];
extern uint32_t _DPF_myIP;
extern ExAddr *_DPF_inagent;
extern char *_DPF_route;

static void
setupEnv()
{
    // Unfortunately DPF does not support multihomed machines.
    // Get my IP address.
    char *addr = getenv("DPF_MY_IP");
    if (addr) {
        struct in_addr inaddr;
        if (!inet_aton(addr, &inaddr)) {
            ABORT("DPF_MY_IP is invalid");
        }
        _DPF_myIP = inaddr.s_addr;
    } else {
        (void)_myIP(&_DPF_myIP);
    }

    // Get inagent. must be of <ip:port> format
    struct sockaddr_in sin;
    addr = getenv("DPF_INAGENT");
    if (addr) {
        if (_str2sin(addr, &sin) < 0) {
            ABORT("DPF_INAGENT contains invalid ip:port");
        }
        _DPF_inagent = new ExAddr(&sin, NULL, NULL);
    }

    // inherit sockets if necessary.
    if (DPF_inheritFrom() < 0) {
        ABORT("failed to inherit sockets");
    }

    // Open hosts file
    char *path = getenv("DPF_ROUTE");
    if (path != NULL) {
        _DPF_route = strdup(path);
    }
}

extern "C" {
void
DPF_init(char *debug_file_path, char *debug_level, int max_debug_file_size)
{
    int i;
    struct sigaction act;

    char *debug_file = NULL;
    if (debug_file_path) {
        debug_file = debug_file_path;
    } else {
        debug_file = getenv("DPF_DEBUG_PATH");
        //if (!debug_file) { ABORT("DPF_DEBUG_PATH undefined"); }
    }
    char *level = NULL;
    if (debug_level) {
        level = debug_level;
    } else {
        level = getenv("DPF_DEBUG_LEVEL");
        //if (!level) { ABORT("DPF_DEBUG_LEVEL undefined"); }
    }
    int dlevel = D_BASIC;
    if (level && !strcasecmp(level, "fulldebug")) {
        dlevel = D_FULL;
    }
    int maxsize;
    if (max_debug_file_size > 0) {
        maxsize = max_debug_file_size;
    } else {
        char *sizestr = getenv("DPF_LOG_MAXSIZE");
        if (sizestr) {
            maxsize = atol(sizestr);
        } else {
            maxsize = 64000;
        }
    }

    Generic_logInit(debug_file, dlevel, maxsize);

    LOG(D_FULL, "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    LOG(D_FULL, "                   DPF_init() called\n");
    // Initialize Data Structure
    for (i = 0; i <= DPF_OPEN_MAX; i++) {
        _DPF_fd[i].type = NOT_DEFINED;
        _DPF_fd[i].info = NULL;
    }
    // Let the process ignore SIGPIPE
    act.sa_handler = SIG_IGN;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if (sigaction(SIGPIPE, &act, NULL) < 0) {
        ABORT("Failed sigaction - %s\n", strerror(errno));
    }
    LOG(D_FULL, "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");

    LOG(D_FULL, "---------------------------------------------------------------\n");
    LOG(D_FULL, "Initial Setup\n");
    setupEnv();
    LOG(D_FULL, "Initial Setup Done\n");
    LOG(D_FULL, "---------------------------------------------------------------\n");
    _DPF_initialized = 1;

    return;
}
} // extern "C"
