#include <sys/socket.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

// solaris
#include <sys/types.h>
#include <sys/socket.h>
// solaris

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {
int
DPF_listen(int fd, int backlog)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    if (_DPF_fd[fd].type == NOT_DEFINED) {
        LOG(D_FULL, "listen($d[NOT_DEFINED])\n", fd);
        return listen(fd, backlog);
    } else if (_DPF_fd[fd].type != TCP_UNCLASSIFIED) {
        LOG(D_BASIC, "DPF_listen called for invalid fd\n");
        errno = EINVAL;
        return -1;
    }

    LOG(D_FULL, "DPF_listen(%d, %d)]\n", fd, backlog);
    DPF_TCP_sock *sock = (DPF_TCP_sock *)_DPF_fd[fd].info;
    /* socket should have been bounded */
    switch(sock->status) {
        case LOCAL_BOUND:
            if (listen(fd, backlog) < 0) {
                LOG(D_BASIC, "local_listen: listen failed\n");
                return -1;
            }
            _DPF_fd[fd].type = TCP_SERVER;
            sock->app_status = LISTENING;
            sock->status = LOCAL_LISTEN;
            return 0;
        case FULL_BOUND:
            break;
        default:
            LOG(D_BASIC, "DPF_listen: socket not bound\n");
            errno = EINVAL;
            return -1;
    }

    if (listen(fd, backlog) < 0) {
        LOG(D_BASIC, "local_listen: listen failed\n");
        _DPF_fd[sock->hfd].type = NOT_DEFINED;
        close(sock->hfd);
        sock->hfd = -1;
        return -1;
    }
    _DPF_fd[fd].type = TCP_SERVER;
    sock->status = LISTENING_1;
    sock->entered = time(NULL);
    TCP_SERV_MNG_state_machine(sock, 0);
    if (sock->status == LISTEN_FAIL) {
        _close_duped_fd(sock->inagent->get_fd());
        delete sock->inagent;
        sock->inagent = NULL;
        LOG(D_BASIC, "Making the socket FULL_LISTEN failed.\n");
        _DPF_fd[sock->hfd].type = NOT_DEFINED;
        close(sock->hfd);
        sock->hfd = -1;
        return -1;
    }
    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->status == FULL_LISTEN) {
            _close_duped_fd(sock->inagent->get_fd());
            delete sock->inagent;
            sock->inagent = NULL;
            sock->app_status = LISTENING;
            return 0;
        } else if (sock->status == LISTEN_FAIL) {
            _close_duped_fd(sock->inagent->get_fd());
            delete sock->inagent;
            sock->inagent = NULL;
            _DPF_fd[sock->hfd].type = NOT_DEFINED;
            close(sock->hfd);
            sock->hfd = -1;
            LOG(D_BASIC, "Making the socket FULL_LISTEN failed.\n");
            return -1;
        }
    }
}
} // extern "C"
