#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {
int
DPF_recvfrom(int fd, void *buf, size_t size, int flags,
        struct sockaddr *from, socklen_t *fromlen)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // check fd
    if (_DPF_fd[fd].type == NOT_DEFINED) {
        return recvfrom(fd, buf, size, flags, from, fromlen);
    } else if (_DPF_fd[fd].type != UDP) {
        LOG(D_BASIC, "DPF_recvfrom called for invalid fd\n");
        errno = EBADF;
        return -1;
    }

    DPF_UDP_sock *sock = (DPF_UDP_sock *)_DPF_fd[fd].info;

    // check status
    if (sock->status != LOCAL_BOUND &&
            sock->status != FULL_BOUND &&
            sock->myaddr == NULL) {
        LOG(D_BASIC, "DPF_recvfrom called for fd not in valid status\n");
        errno = EINVAL;
        return -1;
    }

    if (sock->nonblock || (flags & MSG_DONTWAIT)) {
        LOG(D_FULL, "DPF_recvfrom(fd=%d, buf, %d, nonblock, &from)\n", fd, size);
    } else {
        LOG(D_FULL, "DPF_recvfrom(fd=%d, buf, %d, block, &from)\n", fd, size);
    }

    // if recv buffer is not empty
    if (sock->rbuf.no_msg != 0) {
        unsigned short size = sock->dequeue((struct sockaddr_in *)from, buf);
        LOG(D_FULL, "returning %d bytes from %s\n",
                size, _sin2str((struct sockaddr_in *)from));
        return size;
    } else if (sock->nonblock || (flags & MSG_DONTWAIT)) { // nonblocking
        errno = EAGAIN;
        LOG(D_FULL, "returning EAGAIN\n");
        return -1;
    }

    // here we need to wait until something happen
    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->rbuf.no_msg != 0) { // recv buffer is not empty
            // dequeue a msg
            int size = sock->dequeue((struct sockaddr_in *)from, buf);
            LOG(D_FULL, "returning %d bytes from %s\n",
                    size, _sin2str((struct sockaddr_in *)from));
            return size;
        }
    }
}

} // extern "C"
