#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#include "DPF_client.h"

extern char *_DPF_route;
static unsigned int masks[33] = {
	0x00000000,
	0x80000000, 0xC0000000, 0xE0000000, 0xF0000000,
	0xF8000000, 0xFC000000, 0xFE000000, 0xFF000000,
	0xFF800000, 0xFFC00000, 0xFFE00000, 0xFFF00000,
	0xFFF80000, 0xFFFC0000, 0xFFFE0000, 0xFFFF0000,
	0xFFFF8000, 0xFFFFC000, 0xFFFFE000, 0xFFFFF000,
	0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00, 0xFFFFFF00,
	0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0, 0xFFFFFFF0,
	0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE, 0xFFFFFFFF
};

// @arg:
//   - peer_ip: ip address (in network order) of the peer
//   - DPF: return true if DPF mechanism must be used to contact 'peer_ip'
//   - inagent: The address of the inagent of 'peer_ip' is returned via this
//              argument. If 'DPF' returns false, then this arg must be ignored
// @ret: true, no error
//       false, if error
bool
_DPF_get_route(uint32_t peer_ip, bool *DPF, struct sockaddr_in *inagent)
{
	char line[100], peer[40], agent[40];
    char *str_ip, *str_mask, *next;
	struct in_addr inaddr;
    int bits;
	uint32_t mask;

    // open the routing table
    FILE *rFile = fopen(_DPF_route, "r");
    if (rFile == NULL) {
        ABORT("_DPF_route file (%s) open fail", _DPF_route);
    }

	// File format of DPF routing table:
	//   - each line contains one routing rule
	//   - each routing rule has the following fields:
	//     + peer: ip/mask, where ip is a dotted notation and mask is the number
    //             of bits to be matched
	//     + inagent: ip:port of peer's inagent. If direct communication is required,
    //                then this field is "none"
    while(true) {
        if (fgets(line, 100, rFile) == NULL) {
            if (errno == EAGAIN || errno == EINTR) {
                continue;
	        } else if (feof(rFile)) {
                break;
	        }
		    LOG(D_BASIC, "GCB routing table read failed: %s\n", strerror(errno));
            goto error;
        }
        if (line[0] == '#' || isspace(line[0])) { continue; }
        if (sscanf(line, "%s %s", peer, agent) != 2) {
	        LOG(D_BASIC, "routing table format error: %s\n", line);
            goto error;
        }

		str_ip = peer;
		next = strchr(peer, '/');
		if (next) { // ip/mask format
            *next = '\0';
            str_mask = next + 1;
            bits = atoi(str_mask);
            if (bits < 0 || bits > 32) {
			    LOG(D_BASIC, "invalid network mask\n");
                goto error;
            }
		} else { // ip without mask
            bits = 32;
        }
        // ip
		if (!inet_aton(str_ip, &inaddr)) {
			LOG(D_BASIC, "invalid ipaddr\n");
			goto error;
		}
        // mask
		mask = masks[bits];
		// check if found
		if ((ntohl(inaddr.s_addr) & mask) == (ntohl(peer_ip) & mask)) {
			goto found;
		}
	}
	// Not found. Default: no DPF but outagent needs to be contact
    *DPF = false;
    fclose(rFile);
	return true;

found:
    // DPF & inagent
	if (!strncasecmp(agent, "none", 4)) {
        *DPF = false;
	} else {
        *DPF = true;
        if (_str2sin(agent, inagent) < 0) {
            LOG(D_BASIC, "inagent is invalid format\n");
            goto error;
        }
	}
    fclose(rFile);
    return true;

error:
    fclose(rFile);
    return false;
}
