#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern "C" {

// Global data structure
extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

int
DPF_socket(int domain, int type, int protocol)
{
    int fd;

    // Initialize global data structure if necessary
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // Check if the socket is in the family and type that GCB supports
    if (domain != AF_INET || (type != SOCK_STREAM && type != SOCK_DGRAM)) {
        fd = socket(domain, type, protocol);
        LOG(D_FULL, "socket(%d[NOT_DEFINED])\n", fd);
        if (_DPF_fd[fd].info) {
            LOG(D_BASIC, "DPF_fdInfo[%d] is not clean\n", fd);
            delete _DPF_fd[fd].info;
            _DPF_fd[fd].type = NOT_DEFINED;
            _DPF_fd[fd].info = NULL;
        }
        return fd;
    }

    if (type == SOCK_STREAM) {
        // create nonblocking socket
        if ((fd = _tcp_nonblock_socket()) < 0) {
            LOG(D_BASIC, "DPF_socket: socket creation failed\n");
            return -1;
        }
        LOG(D_FULL, "[DPF_socket(fd = %d, TCP)]\n", fd);
        // Register the sock as TCP_UNCLASSIFIED
        DPF_TCP_sock *tcp_sock = new DPF_TCP_sock(fd);
        _register_sock(fd, TCP_UNCLASSIFIED, tcp_sock);
    } else if (type == SOCK_DGRAM) {
        // create nonblocking socket
        if ((fd = _udp_nonblock_socket()) < 0) {
            LOG(D_BASIC, "DPF_socket: socket creation failed\n");
            return -1;
        }
        LOG(D_FULL, "[DPF_socket(fd = %d, UDP)]\n", fd);
        DPF_UDP_sock *udp_sock = new DPF_UDP_sock(fd);
        _register_sock(fd, UDP, udp_sock);
    } else {
        LOG(D_BASIC, "DPF_socket: invalid protocol passed\n");
        errno = EINVAL;
        return -1;
    }

    return fd;
}

}
