#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>

#include "DPF_common.h"

char *_fd_type_str[] = {
    "NOT_DEFINED", "TCP_UNCLASSIFIED", "TCP_SERVER", "TCP_CLIENT", "UDP",
    "TCP_IN_MNGER", "TCP_OUT_MNGER", "TCP_PEER_IN_MNGER",
    "UDP_IN_MNGER", "UDP_OUT_MNGER", "UDP_PEER_IN_MNGER",
    "HIDDEN", "ACCEPTED",
    "LISTENER", "CONNECTOR", "TCP_RELAY", "UDP_RELAY",
    "HEARTBEAT_FD",
    "RESERVED"
};
char *_msg_type_str[] = {
    "DPF_CMD", "CLIENT_HELLO", "SERVER_HELLO", "CLIENT_CERTIFICATE",
    "SERVER_CERTIFICATE", "KEY_EXCHANGE", "KEY_ACK", "CLIENT_FINISHED", "SERVER_FINISHED"
};
char *_CMD_opcode_str[] = {
    "TCP_BIND", "UDP_BIND", "TCP_OPEN", "UDP_OPEN", "TCP_OPEN_DEL", "UDP_OPEN_DEL",
    "LISTEN", "TCP_BIND_OK", "UDP_BIND_OK",
    "TCP_OPEN_OK", "UDP_OPEN_OK", "CONTACT", "TRIGGER", "RELAY_CONTACT", "RELAY_TRIGGER",
    "LISTEN_OK", "TCP_CONNECT", "UDP_CONNECT", "ACTIVE_CONNECT", "ACTIVE_INIT",
    "RELAY_CONNECT", "RELAY_INIT", "PASSIVE_CONNECT", "PASSIVE_INIT", "PASSIVE_ACK",
    "NAK", "DUMMY", "DUMMY_ACK", "TCP_HEARTBEAT", "UDP_HEARTBEAT"
};

unsigned short
_stripe_UDP_msg(char *buf,
        DPF_CMD_opcode opcode,
        struct sockaddr_in *local,
        struct sockaddr_in *proxy)
{
    char str[100];

    buf[0] = DPF_TAG;
    buf[1] = DPF_VERSION;
    uint16_t nshort = htons(18);
    memcpy(&buf[2], &nshort, 2);
    buf[4] = DPF_CMD;
    buf[5] = opcode;
    memcpy(&buf[6], &(local->sin_addr.s_addr), 4);
    memcpy(&buf[10], &(local->sin_port), 2);
    sprintf(str, "local=%s ", _sin2str(local));
    memcpy(&buf[12], &(proxy->sin_addr.s_addr), 4);
    memcpy(&buf[16], &(proxy->sin_port), 2);
    sprintf(&str[strlen(str)], "proxy=%s", _sin2str(proxy));

    switch(opcode) {
        case TCP_OPEN_DEL:
            LOG(D_FULL, ">> TCP_OPEN_DEL: %s\n", str);
            break;
        case UDP_OPEN_DEL:
            LOG(D_FULL, ">> UDP_OPEN_DEL: %s\n", str);
            break;
        case TCP_HEARTBEAT:
            LOG(D_FULL, ">> TCP_HEARTBEAT: %s\n", str);
            break;
        case UDP_HEARTBEAT:
            LOG(D_FULL, ">> UDP_HEARTBEAT: %s\n", str);
            break;
        default:
            ABORT("Invalid opcode passed to _stripe_UDP_msg");
    }
    return 18;
}

bool
_unstripe_UDP_msg(char *buf, unsigned short buf_len, DPF_CMD_opcode *opcode,
        struct sockaddr_in *addr1, struct sockaddr_in *addr2, char *byte)
{
    if ((unsigned char)buf[0] != DPF_TAG ||
        (unsigned char)buf[1] != DPF_VERSION ||
        (DPF_msg_type)buf[4] != DPF_CMD)
    {
        LOG(D_BASIC, "header check failed\n");
        return false;
    }
    uint16_t nshort;
    memcpy(&nshort, &buf[2], 2);
    if (ntohs(nshort) != buf_len) {
        LOG(D_BASIC, "length field check failed\n");
        return false;
    }
    char str[100];
    char *body = &buf[5];
    *opcode = (DPF_CMD_opcode) body[0];
    sprintf(str, "%s ", _CMD_opcode_str[*opcode]);
    memcpy(&(addr1->sin_addr.s_addr), &body[1], 4);
    memcpy(&(addr1->sin_port), &body[5], 2);
    sprintf(&str[strlen(str)], "local = %s ", _sin2str(addr1));
    memcpy(&(addr2->sin_addr.s_addr), &body[7], 4);
    memcpy(&(addr2->sin_port), &body[11], 2);
    sprintf(&str[strlen(str)], "proxy = %s ", _sin2str(addr2));

    LOG(D_FULL, "<< DPF_CMD(UDP): %s\n", str);

    return true;
}
