#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#include "DPF_common.h"

ExAddr::ExAddr(
        struct sockaddr_in *addr1,
        struct sockaddr_in *addr2,
        struct sockaddr_in *addr3)
{
    if (addr1) {
        local = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
        if (!local) { ABORT("malloc failed"); }
        memcpy(local, addr1, sizeof(struct sockaddr_in));
    } else {
        local = NULL;
    }
    if (addr2) {
        proxy = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
        if (!proxy) { ABORT("malloc failed"); }
        memcpy(proxy, addr2, sizeof(struct sockaddr_in));
    } else {
        proxy = NULL;
    }
    if (addr3) {
        agent = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
        if (!agent) { ABORT("malloc failed"); }
        memcpy(agent, addr3, sizeof(struct sockaddr_in));
    } else {
        agent = NULL;
    }
}

ExAddr::~ExAddr()
{
    if (local) { free(local); }
    if (proxy) { free(proxy); }
    if (agent) { free(agent); }
}

void
ExAddr::set_local(struct sockaddr_in *addr)
{
    if (addr) {
        if (!local) {
            local = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
            if (!local) { ABORT("malloc failed"); }
        }
        memcpy(local, addr, sizeof(struct sockaddr_in));
    } else if (local) {
        free(local);
        local = NULL;
    }
}

void
ExAddr::set_proxy(struct sockaddr_in *addr)
{
    if (addr) {
        if (!proxy) {
            proxy = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
            if (!proxy) { ABORT("malloc failed"); }
        }
        memcpy(proxy, addr, sizeof(struct sockaddr_in));
    } else if (proxy) {
        free(proxy);
        proxy = NULL;
    }
}

void
ExAddr::set_agent(struct sockaddr_in *addr)
{
    if (addr) {
        if (!agent) {
            agent = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
            if (!agent) { ABORT("malloc failed"); }
        }
        memcpy(this->agent, addr, sizeof(struct sockaddr_in));
    } else if (agent) {
        free(agent);
        agent = NULL;
    }
}

bool
ExAddr::same_proxy(ExAddr *addr)
{
    if (proxy && addr->proxy &&
        proxy->sin_addr.s_addr == addr->proxy->sin_addr.s_addr &&
        proxy->sin_port == addr->proxy->sin_port) {
        return true;
    }
    return false;
}

bool
ExAddr::same_proxy(struct sockaddr_in *sin)
{
    if (proxy &&
        proxy->sin_addr.s_addr == sin->sin_addr.s_addr &&
        proxy->sin_port == sin->sin_port)
    { return true; }

    return false;
}

int
ExAddr::connect(int fd)
{
    if (local == NULL) {
        LOG(D_BASIC, "ExAddr::connect is called when local = NULL\n");
        errno = EINVAL;
        return -1;
    }
    return ::connect(fd, (struct sockaddr *)local, sizeof(struct sockaddr_in));
}

int
ExAddr::connect_agent(int fd)
{
    if (agent == NULL) {
        LOG(D_BASIC, "ExAddr::connect_agent is called when agent = NULL\n");
        errno = EINVAL;
        return -1;
    }
    return ::connect(fd, (struct sockaddr *)agent, sizeof(struct sockaddr_in));
}

void
ExAddr::tostr(const char *prefix, char *str)
{
    if (proxy) {
        sprintf(str, "%sproxy=%s ", prefix, _sin2str(proxy));
    } else {
        sprintf(str, "%sproxy=NULL ", prefix);
    }
    if (local) {
        sprintf(&str[strlen(str)], "local=%s ", _sin2str(local));
    } else {
        sprintf(&str[strlen(str)], "local=NULL ");
    }
    if (agent) {
        sprintf(&str[strlen(str)], "agent=%s ", _sin2str(agent));
    } else {
        sprintf(&str[strlen(str)], "agent=NULL ");
    }
}

int
ExAddr::stripe(char *buf)
{
    // the size of the serialization excluding this field(2 bytes)
    // tag if local is not NULL(1 byte)
    // local: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    // tag if proxy is not NULL(1 byte)
    // proxy: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    // tag if agent is not NULL(1 byte)
    // agent: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)

    char str[400];
    int index = 2;
    // tag if local is not NULL(1 byte)
    // local: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    if (local) {
        sprintf(str, "[%d]yes", index);
        buf[index++] = 1;
        memcpy(&buf[index], local, sizeof(struct sockaddr_in));
        sprintf(&str[strlen(str)], "[%d]local=%s ", index, _sin2str(local));
        index += sizeof(struct sockaddr_in);
    } else {
        sprintf(str, "[%d]no", index);
        buf[index++] = 0;
    }
    // tag if proxy is not NULL(1 byte)
    // proxy: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    if (proxy) {
        sprintf(&str[strlen(str)], "[%d]yes", index);
        buf[index++] = 1;
        memcpy(&buf[index], proxy, sizeof(struct sockaddr_in));
        sprintf(&str[strlen(str)], "[%d]proxy=%s ", index, _sin2str(local));
        index += sizeof(struct sockaddr_in);
    } else {
        sprintf(&str[strlen(str)], "[%d]no", index);
        buf[index++] = 0;
    }
    // tag if agent is not NULL(1 byte)
    // agent: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    if (agent) {
        sprintf(&str[strlen(str)], "[%d]yes", index);
        buf[index++] = 1;
        memcpy(&buf[index], agent, sizeof(struct sockaddr_in));
        sprintf(&str[strlen(str)], "[%d]agent=%s ", index, _sin2str(local));
        index += sizeof(struct sockaddr_in);
    } else {
        sprintf(&str[strlen(str)], "[%d]no", index);
        buf[index++] = 0;
    }

    // the size of the serialization excluding this field(2 bytes)
    uint16_t nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of ExAddr: [0]len=%d%s\n", index-2, str);

    return index;
}

int
ExAddr::unstripe(char *buf)
{
    // the size of the serialization excluding this field(2 bytes)
    // tag if local is not NULL(1 byte)
    // local: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    // tag if proxy is not NULL(1 byte)
    // proxy: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    // tag if agent is not NULL(1 byte)
    // agent: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)

    char str[400];
    int index = 0;
    uint16_t nshort;
    // the size of the serialization excluding this field(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // tag if local is not NULL(1 byte)
    // local: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    if (buf[index++] == 1) {
        sprintf(&str[strlen(str)], "[%d]yes", index-1);
        if (!local) {
            local = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
            if (!local) { ABORT("malloc failed"); }
        }
        memcpy(local, &buf[index], sizeof(struct sockaddr_in));
        sprintf(&str[strlen(str)], "[%d]local=%s ", index, _sin2str(local));
        index += sizeof(struct sockaddr_in);
    } else {
        sprintf(&str[strlen(str)], "[%d]no", index-1);
    }
    // tag if proxy is not NULL(1 byte)
    // proxy: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    if (buf[index++] == 1) {
        sprintf(&str[strlen(str)], "[%d]yes", index-1);
        if (!proxy) {
            proxy = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
            if (!proxy) { ABORT("malloc failed"); }
        }
        memcpy(proxy, &buf[index], sizeof(struct sockaddr_in));
        sprintf(&str[strlen(str)], "[%d]proxy=%s ", index, _sin2str(local));
        index += sizeof(struct sockaddr_in);
    } else {
        sprintf(&str[strlen(str)], "[%d]no", index-1);
    }
    // tag if agent is not NULL(1 byte)
    // agent: present only if the tag above is 1(sizeof(struct sockaddr_in) bytes)
    if (buf[index++] == 1) {
        sprintf(&str[strlen(str)], "[%d]yes", index-1);
        if (!agent) {
            agent = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
            if (!agent) { ABORT("malloc failed"); }
        }
        memcpy(agent, &buf[index], sizeof(struct sockaddr_in));
        sprintf(&str[strlen(str)], "[%d]agent=%s ", index, _sin2str(local));
        index += sizeof(struct sockaddr_in);
    } else {
        sprintf(&str[strlen(str)], "[%d]no", index-1);
    }

    LOG(D_FULL, "Unserialization of ExAddr: %s\n", str);

    if (length != index-2) {
        ABORT("length does not match with the actual size of the serialization");
    }

    return index;
}
