#ifndef __FW_MNGER
#define __FW_MNGER

#include <stdio.h>
#include <netinet/in.h>

class FW_Mnger {
	public:
		// @args: proto: TCP or UDP
        FW_Mnger(char proto, char *logfile);

        // Create a pinhole for the pair of ("src", "dst"). Either but not both
        // can be NULL, meaning anybody.
        bool open(struct sockaddr_in *src, struct sockaddr_in *dst);

        // Delete the pinhole for the pair of ("src", "dst"). Either but not both
        // can be NULL, meaning anybody.
        bool close(struct sockaddr_in *src, struct sockaddr_in *dst);

        // At the startup time, the log file is refered to delete rules that are left
        // behind the previous crash. To maintain the log file not too big, the log
        // file must be periodically (once a day at least) compressed. Actually no
        // compression occurrs here. This routine makes the log file compact by
        // deleting unnecessary entries in the file
        void compress_log();

        // Delete rules remained in the log file and start with a fresh log file.
        // This routine must be called at the startup time
        void cleanup();

    private:
        static const char TCP = 't';
        static const char UDP = 'u';
        static const int HASH_SIZE = 173;

	protected:
		int _protocol;
		char _logFileName[30];
        FILE *_logFD;
        /* Local representation of firewall rules. Each rule is
         * represented by (src ip, src port, local ip, local port, dest ip, dest port).
         * (local ip, local port) is not used here. This is for NAT manager
         */
		struct fwRule *_rules[HASH_SIZE];

		/* Method */
        // Add a rule to the internal data structure. A rule is internally represented by
        // (private address, local address, remote address), where pviate address is ip:port
        // of a socket behind firewall, local address is ip:port that a NAT binding will
        // be created for the private address, and remote address is ip:port of a socket
        // outside the private/firewalled network. For firewall management, we use private
        // address as the source address and remote address as the destination address.
        // @return: true, if success
        //          false if error
        // @error:
        //    - EINVAL: the rule with the same value already exists
		bool addInternal(uint32_t src_ip, uint16_t src_port,
                uint32_t dst_ip, uint16_t dst_port);
        // Delete a rule from the internal representation. See the comments on addInternal
        // for how rules are internally represented
        // @return: true, if success
        //          false if error
        // @error:
        //    - EINVAL: the rule with the given value does not exist
        bool deleteInternal(uint32_t src_ip, uint16_t src_port,
                uint32_t dst_ip, uint16_t dst_port);
		// Get the list of rules which are in the local representation.
		// Note: The list of rules should be free by the caller.
		struct fwRule *getRules();
        // Delete all the rules from the local representation
        void flush_internal();
};

#endif
