#ifndef _FWKERNER_H
#define _FWKERNER_H

// Append or delete a DNAT rule. A rule is denoted by
// "{sip:sport->lip:lport}-->{sip:sport -> dip:dport}", meaning that packets from
// sip:sport to lip:lport are rewritten as if from sip:sport to dip:dport. lip:lport
// must be a valid public address in this machine and dip:dport be a valid address
// in a machine behind this machine
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - command: CMD_ADD or CMD_DEL
//  - sip, sport, lip, lport, dip, dport: in network byte order
int dnatRuleSet(int protocol, int command,
        uint32_t sip, uint16_t sport,
        uint32_t lip, uint16_t lport,
        uint32_t dip, uint16_t dport);

// List DNAT rules. A rule is denoted by
// "{sip:sport->lip:lport}-->{sip:sport -> dip:dport}", meaning that packets from
// sip:sport to lip:lport are rewritten as if from sip:sport to dip:dport. lip:lport
// must be a valid public address in this machine and dip:dport be a valid address
// in a machine behind this machine
int dnatRuleList(int protocol, struct fwRule **rules);

// Append or delete a SNAT rule. A rule is denoted by
// "{sip:sport->dip:dport}-->{lip:lport->dip:dport}", meaning that packets from
// sip:sport to dip:dport are rewritten as if from lip:lport to dip:dport.
// lip:lport must be a valid public address in this machine and sip:sport be a valid
// address in a machine behind this machine
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - command: CMD_ADD or CMD_DEL
//  - sip, sport, lip, lport, dip, dport: in network byte order
int snatRuleSet(int protocol, int command,
        uint32_t sip, uint16_t sport,
        uint32_t lip, uint16_t lport,
        uint32_t dip, uint16_t dport);

// List SNAT rules. A rule is denoted by
// "{sip:sport->dip:dport}-->{lip:lport->dip:dport}", meaning that packets from
// sip:sport to dip:dport are rewritten as if from lip:lport to dip:dport.
// lip:lport must be a valid public address in this machine and sip:sport be a valid
// address in a machine behind this machine
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - rules: the list is returned via this arg
int snatRuleList(int protocol, struct fwRule **rules);

// Append or delete a filter rule. A rule is denoted by
// "{sip:sport}-->{dip:dport}", meaning that packets from
// sip:sport to dip:dport must be accepted or rejected
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - command: CMD_ADD or CMD_DEL
//  - sip, sport, dip, dport: in network byte order
int filterRuleSet(int protocol, int command,
        uint32_t sip, uint16_t sport,
        uint32_t dip, uint16_t dport);

// List filter rules. A rule is denoted by
// "{sip:sport}-->{dip:dport}", meaning that packets from
// sip:sport to dip:dport are accepted or rejected
int filterRuleList(int protocol, struct fwRule **rules);

/* Returns the error string which explains the most recent error */
char *strError(void);
#endif
