#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "inagent.h"

extern char *_DPF_sock_stat_str[];

static int
max(int a, int b, int c, int d)
{
    if (a > b) {
        if (a > c) {
            if (a > d) { return a; }
            else { return d; }
        } else {
            if (c > d) { return c; }
            else { return d; }
        }
    } else {
        if (b > c) {
            if (b > d) { return b; }
            else { return d; }
        } else {
            if (c > d) { return c; }
            else { return d; }
        }
    }
}

// Constructor for firewall manager
InAgent::InAgent(
        struct sockaddr_in *servaddr,
        int tcp_serverfd,
        int udp_serverfd,
        int time_rfd)
{
    _firewall = true;
    _TCP_free_port = NULL;
    _UDP_free_port = NULL;
    _TCP_NAT_mnger = NULL;
    _UDP_NAT_mnger = NULL;
    //_proxyaddr = undefined
    //_clientaddr = undefined
    _clientfd = -1;

    memcpy(&_servaddr, servaddr, sizeof(_servaddr));

    for (int i = 0; i <= DPF_OPEN_MAX; i++) {
        _client_fd[i].type = NOT_DEFINED;
        _client_fd[i].info = NULL;
    }
    _tcp_servfd = tcp_serverfd;
    _client_fd[_tcp_servfd].type = RESERVED;
    _udp_servfd = udp_serverfd;
    _client_fd[_udp_servfd].type = RESERVED;
    _timeout = time_rfd;
    _client_fd[_timeout].type = RESERVED;
    _maxfd = max(_tcp_servfd, _udp_servfd, _clientfd, _timeout) + 1;

    for (int i=0; i<DPF_HASH_SIZE; i++) {
        _proxy_hash[i] = NULL;
    }

    _TCP_FW_mnger = new FW_Mnger(TCP, ".filter_log.tcp");
    if (!_TCP_FW_mnger) {
        ABORT("Couldn't create _TCP_FW_mnger object");
    }
    _UDP_FW_mnger = new FW_Mnger(UDP, ".filter_log.udp");
    if (!_UDP_FW_mnger) {
        ABORT("Couldn't create _UDP_FW_mnger object");
    }
    //dprintf(D_ALWAYS, "Initial cleanup...\n");
    _TCP_FW_mnger->cleanup();
    _UDP_FW_mnger->cleanup();

    LOG(D_BASIC, "=================================================\n");
    LOG(D_BASIC, "\t_servfd[tcp=%d,udp=%d]: %s\n", _tcp_servfd, _udp_servfd, _sin2str(&_servaddr));
    LOG(D_BASIC, "\t_timeout[fd=%d]\n", _timeout);
    LOG(D_BASIC, "=================================================\n");
}

// Constructor for NAT manager
InAgent::InAgent(
        struct sockaddr_in *proxyaddr,
        struct sockaddr_in *servaddr,
        struct sockaddr_in *clientaddr,
        int tcp_serverfd,
        int udp_serverfd,
        int clientfd,
        int time_rfd)
{
    _firewall = false;
    _TCP_FW_mnger = NULL;
    _UDP_FW_mnger = NULL;

    memcpy(&_proxyaddr, proxyaddr, sizeof(_proxyaddr));
    memcpy(&_servaddr, servaddr, sizeof(_servaddr));
    memcpy(&_clientaddr, clientaddr, sizeof(_clientaddr));

    for (int i = 0; i <= DPF_OPEN_MAX; i++) {
        _client_fd[i].type = NOT_DEFINED;
        _client_fd[i].info = NULL;
    }
    _tcp_servfd = tcp_serverfd;
    _client_fd[_tcp_servfd].type = RESERVED;
    _udp_servfd = udp_serverfd;
    _client_fd[_udp_servfd].type = RESERVED;
    _clientfd = clientfd;
    _client_fd[_clientfd].type = RESERVED;
    _timeout = time_rfd;
    _client_fd[_timeout].type = RESERVED;
    _maxfd = max(_tcp_servfd, _udp_servfd, _clientfd, _timeout) + 1;

    for (int i=0; i<DPF_HASH_SIZE; i++) {
        _proxy_hash[i] = NULL;
    }

    _TCP_free_port = new FreePortMnger();
    if (!_TCP_free_port) {
        ABORT("Couldn't create _TCP_free_port object");
    }
    _TCP_free_port->addInterface(proxyaddr->sin_addr.s_addr);
    _UDP_free_port = new FreePortMnger();
    if (!_UDP_free_port) {
        ABORT("Couldn't create _UDP_free_port object");
    }
    _UDP_free_port->addInterface(proxyaddr->sin_addr.s_addr);
    _TCP_NAT_mnger = new NAT_Mnger(TCP, ".inagent_dnat_log.tcp");
    if (!_TCP_NAT_mnger) {
        ABORT("Couldn't create _TCP_NAT_mnger object");
    }
    _UDP_NAT_mnger = new NAT_Mnger(UDP, ".inagent_dnat_log.udp");
    if (!_UDP_NAT_mnger) {
        ABORT("Couldn't create _UDP_NAT_mnger object");
    }
    //dprintf(D_ALWAYS, "Initial cleanup...\n");
    _TCP_NAT_mnger->cleanup();
    _UDP_NAT_mnger->cleanup();

    LOG(D_BASIC, "=================================================\n");
    LOG(D_BASIC, "\t_servfd[tcp=%d,udp=%d]: %s\n", _tcp_servfd, _udp_servfd, _sin2str(&_servaddr));
    LOG(D_BASIC, "\t_clientfd[fd=%d]: %s\n", _clientfd, _sin2str(&_clientaddr));
    LOG(D_BASIC, "\t_timeout[fd=%d]\n", _timeout);
    LOG(D_BASIC, "=================================================\n");
}

InAgent::~InAgent()
{
    if (_TCP_FW_mnger) { delete _TCP_FW_mnger; }
    if (_UDP_FW_mnger) { delete _UDP_FW_mnger; }
    if (_TCP_free_port) { delete _TCP_free_port; }
    if (_UDP_free_port) { delete _UDP_free_port; }
    if (_TCP_NAT_mnger) { delete _TCP_NAT_mnger; }
    if (_UDP_NAT_mnger) { delete _UDP_NAT_mnger; }
}

// Add a fresh listener. We don't know the listener's address yet.
void
InAgent::add_listener()
{
    struct sockaddr_in sin;
    socklen_t addrlen = sizeof(sin);

    int fd = accept(_tcp_servfd, (struct sockaddr *)&sin, &addrlen);
    if (fd < 0) {
        LOG(D_BASIC, "accept failed: %s\n", strerror(errno));
        return;
    }
    if (_make_nonblock(fd) < 0) {
        LOG(D_BASIC, "can't make fd nonblock\n");
        close(fd);
    }

    Listener *listener = new Listener(fd);

    if (fd >= _maxfd) {
        _maxfd = fd + 1;
    }
    _client_fd[fd].type = LISTENER;
    _client_fd[fd].info = (void *)listener;

    LOG(D_FULL, "New listener created: fd[%d] connection accepted from %s\n",
            fd, _sin2str(&sin));
    return;
}

// Add a listener who we lost for some reason. There are plenty of chance that
// listener's information get deleted. Few examples are 1) inagent has crashed
// 2) a series of heartbeats from the listener, which are UDP packets, might get lost
void
InAgent::add_listener(char proto, struct sockaddr_in *local, struct sockaddr_in *proxy)
{
    Listener *listener = new Listener(-1);
    listener->proto = proto;
    if (_firewall) {
        if (!(firewall_open(listener, local))) {
            LOG(D_BASIC, "Can't open firewall for the listener\n");
            delete_listener(listener);
            return;
        }
        listener->opened = true;
        listener->myaddr = new ExAddr(local, local, NULL);
    } else {
        if (!(nat_bind(listener, local, proxy))) {
            LOG(D_BASIC, "Can't give the address to the listener again\n");
            delete_listener(listener);
            return;
        }
        listener->opened = true;
        listener->myaddr = new ExAddr(local, proxy, NULL);
    }
    listener_2_hash(listener);
    if (proto == TCP) {
        listener->status = FULL_LISTEN;
    } else {
        listener->status = FULL_BOUND;
    }

    LOG(D_FULL, "New listener created: proxy=%s\n", _sin2str(listener->myaddr->proxy));
    return;
}

bool
InAgent::firewall_open(Listener *listener, struct sockaddr_in *dest)
{
    FW_Mnger *fw_mnger;
    if (listener->proto == TCP) {
        fw_mnger = _TCP_FW_mnger;
    } else {
        fw_mnger = _UDP_FW_mnger;
    }
    bool result = fw_mnger->open(NULL, dest);
    if (!result) {
        LOG(D_BASIC, "firewall_open: Can't open the firewall for the listener\n");
        listener->error = EADDRINUSE;
        return false;
    }
    return true;
}

void
InAgent::firewall_close(Listener *listener)
{
    if (listener->proto == TCP) {
        _TCP_FW_mnger->close(NULL, listener->myaddr->proxy);
    } else {
        _UDP_FW_mnger->close(NULL, listener->myaddr->proxy);
    }
}

bool
InAgent::nat_bind(Listener *listener, struct sockaddr_in *local, struct sockaddr_in *proxy)
{
    FreePortMnger *freeport;
    NAT_Mnger *nat_mnger;
    if (listener->proto == TCP) {
        freeport = _TCP_free_port;
        nat_mnger = _TCP_NAT_mnger;
    } else {
        freeport = _UDP_free_port;
        nat_mnger = _UDP_NAT_mnger;
    }
    bool result;
    if (proxy->sin_port != 0) {
        // the listener wants to use a well-known port
        result = freeport->nextFree(proxy->sin_port, &(proxy->sin_addr.s_addr));
        if (!result) {
            LOG(D_BASIC, "nat_bind: address is not available for the given port\n");
            listener->error = EADDRINUSE;
            return false;
        }
        result = nat_mnger->enable_inbound(NULL, proxy, local);
        if (!result) {
            LOG(D_BASIC, "nat_bind: Can't open NAT for the listener\n");
            listener->error = EADDRINUSE;
            return false;
        }
    } else {
        // the listener wants to use a dynamic address
        result = freeport->nextFree(&(proxy->sin_addr.s_addr), &(proxy->sin_port));
        if (!result) {
            LOG(D_BASIC, "nat_bind: address is not available for the given port\n");
            listener->error = EADDRINUSE;
            return false;
        }
        result = nat_mnger->enable_inbound(NULL, proxy, local);
        if (!result) {
            LOG(D_BASIC, "nat_bind: Can't open NAT for the listener\n");
            listener->error = EADDRINUSE;
            return false;
        }
    }
    return true;
}

void
InAgent::nat_close(Listener *listener)
{
    if (listener->proto == TCP) {
        _TCP_free_port->makeFree(listener->myaddr->proxy->sin_addr.s_addr,
                listener->myaddr->proxy->sin_port);
        _TCP_NAT_mnger->disable_inbound(NULL, listener->myaddr->proxy,
                listener->myaddr->local);
    } else {
        _UDP_free_port->makeFree(listener->myaddr->proxy->sin_addr.s_addr,
                listener->myaddr->proxy->sin_port);
        _UDP_NAT_mnger->disable_inbound(NULL, listener->myaddr->proxy,
                listener->myaddr->local);
    }
}

void
InAgent::listener_2_hash(Listener *listener)
{
    unsigned short index = _hash_sin(listener->myaddr->proxy, DPF_HASH_SIZE);
    if (_proxy_hash[index]) {
        listener->next = _proxy_hash[index];
        _proxy_hash[index]->pre = listener;
    } else {
        //listener->next = NULL;
    }
    _proxy_hash[index] = listener;
    //listener->pre = NULL;
}

Listener *
InAgent::find_listener(struct sockaddr_in *addr, char proto)
{
    unsigned short index = _hash_sin(addr, DPF_HASH_SIZE);
    Listener *ptr = _proxy_hash[index];
    while(ptr) {
        if (addr->sin_addr.s_addr == ptr->myaddr->proxy->sin_addr.s_addr &&
            addr->sin_port == ptr->myaddr->proxy->sin_port &&
            proto == ptr->proto)
        {
            if ((proto == TCP && ptr->status == FULL_LISTEN) ||
                (proto == UDP && ptr->status == FULL_BOUND)) {
                return ptr;
            }
        }
        ptr = ptr->next;
    }
    return NULL;
}

void
InAgent::handle_UDP_msg()
{
    char buf[DPF_MAX_CMD_SIZE];
    struct sockaddr_in from;
    socklen_t addrlen;
    while(true) {
        // recv msg and unstripe it
        addrlen = sizeof(from);
        short size = recvfrom(_udp_servfd, buf, DPF_MAX_CMD_SIZE, 0, (struct sockaddr
                    *)&from, &addrlen);
        if (size < 0) {
            if (errno == EINTR || errno == EAGAIN) { return; }
            ABORT("recvfrom(_udp_servfd) failed");
        }
        DPF_CMD_opcode opcode;
        struct sockaddr_in addr1, addr2;
        char byte;
        if (!_unstripe_UDP_msg(buf, size, &opcode, &addr1, &addr2, &byte)) {
            LOG(D_BASIC, "Can't unstripe the UDP msg received\n");
            continue;
        }
        // if HEARTBEAT
            // find the listener and reset missed
        // if DEL
            // find the listener and delete it
        Listener *listener;
        switch(opcode) {
            case TCP_HEARTBEAT:
                // find the listener and reset missed
                listener = find_listener(&addr2, TCP);
                if (listener) {
                    listener->missed = 0;
                } else {
                    // the listener does not exist, create a new listener
                    // at the address
                    add_listener(TCP, &addr1, &addr2);
                }
                break;
            case UDP_HEARTBEAT:
                // find the listener and reset missed
                listener = find_listener(&addr2, UDP);
                if (listener) {
                    listener->missed = 0;
                } else {
                    // the listener does not exist, create a new listener
                    // at the address
                    add_listener(UDP, &addr1, &addr2);
                }
                break;
            case TCP_OPEN_DEL:
                // find the listener and delete it
                listener = find_listener(&addr2, TCP);
                if (listener) {
                    delete_listener(listener);
                }
                break;
            case UDP_OPEN_DEL:
                // find the listener and delete it
                listener = find_listener(&addr2, UDP);
                if (listener) {
                    delete_listener(listener);
                }
                break;
            default:
                LOG(D_BASIC, "strange msg received. Ignore it!\n");
        }
    } // while(true)
}

void
InAgent::delete_connection(Listener *listener)
{
    int fd = listener->conn->get_fd();
    // adjust _maxfd, if necessary
    if (fd == _maxfd - 1) {
        setmaxfd();
    }
    // clear _client_fd[]
    _client_fd[fd].type = NOT_DEFINED;
    _client_fd[fd].info = NULL;
    // close the socket
    close(fd);
    // delete the object
    delete listener->conn;
    listener->conn = NULL;
    return;
}

void
InAgent::delete_listener(Listener *listener)
{
    // unchain from _proxy_hash[]
    if (listener->next) {
        listener->next->pre = listener->pre;
    }
    if (listener->pre) {
        listener->pre->next = listener->next;
    } else if (listener->myaddr && listener->myaddr->proxy) {
        unsigned short index = _hash_sin(listener->myaddr->proxy, DPF_HASH_SIZE);
        if (_proxy_hash[index] == listener) {
            _proxy_hash[index] = listener->next;
        }
    }

    // close firewall/NAT
    if (listener->opened) {
        if (_firewall) {
            firewall_close(listener);
        } else {
            nat_close(listener);
        }
    }

    if (listener->conn) {
        int mfd = listener->conn->get_fd();
        // adjust _maxfd, if necessary
        if (mfd == _maxfd - 1) {
            LOG(D_FULL, "delete_listener: calling setmaxfd with fd = %d\n", mfd);
            setmaxfd();
        }
        // clear _client_fd[]
        _client_fd[mfd].type = NOT_DEFINED;
        _client_fd[mfd].info = NULL;
        close(mfd);
    }

    delete listener;
}

void
InAgent::add_connector()
{
    struct sockaddr_in sin;
    socklen_t addrlen = sizeof(sin);

    int fd = accept(_clientfd, (struct sockaddr *)&sin, &addrlen);
    if (fd < 0) {
        LOG(D_BASIC, "accept failed: %s\n", strerror(errno));
        return;
    }
    if (_make_nonblock(fd) < 0) {
        LOG(D_BASIC, "can't make fd nonblock\n");
        close(fd);
    }

    Connector *connector = new Connector(fd);
    LOG(D_FULL, "CONNECTOR created: fd[%d] %s\n", fd, _sin2str(&sin));

    if (fd >= _maxfd) {
        _maxfd = fd + 1;
    }
    _client_fd[fd].type = CONNECTOR;
    _client_fd[fd].info = (void *)connector;
}

void
InAgent::delete_connector(Connector *connector)
{
    int mfd = connector->conn->get_fd();

    // adjust _maxfd, if necessary
    if (mfd == _maxfd - 1) {
        LOG(D_FULL, "delete_connector: calling setmaxfd with fd = %d\n", mfd);
        setmaxfd();
    }

    // clear _client_fd[]
    _client_fd[mfd].type = NOT_DEFINED;
    _client_fd[mfd].info = NULL;

    close(mfd);

    delete connector;
}

void
InAgent::setmaxfd()
{
    int start = _maxfd - 2;
    for (int i = start; i >= 0; i--) {
        if (_client_fd[i].type != NOT_DEFINED) {
            _maxfd = i + 1;
            LOG(D_FULL, "\t<setmaxfd> _maxfd=%d\n", _maxfd);
            return;
        }
    }
    ABORT("Wierd!!!");
}
