#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#include "log.h"
#include "DPF.h"

// run as "echo_client_1 server_fqdn port"
int
main(int argc, char *argv[])
{
    if (argc != 3) {
        printf("usage: echo_client_1 server_fqdn port\n");
        return -1;
    }

    // create a socket
    int fd = DPF_socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        printf("DPF_socket failed: %s\n", strerror(errno));
        return -1;
    }
    printf("socket(fd=%d) created\n", fd);

    // bind the socket
    struct sockaddr_in myaddr;
    myaddr.sin_family = AF_INET;
    myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    myaddr.sin_port = htons(0);
    if (DPF_bind(fd, (struct sockaddr *)&myaddr, sizeof(myaddr)) < 0) {
        printf("DPF_bind failed: %s\n", strerror(errno));
        return -1;
    }

    // make the socket nonblocking
    int val;
    if ((val = DPF_fcntl(fd, F_GETFL, 0)) < 0) {
        LOG(D_BASIC, "_make_nonblock: fcntl failed - %s\n", strerror(errno));
        return -1;
    }
    if (DPF_fcntl(fd, F_SETFL, (void *)(val | O_NONBLOCK)) < 0) {
        LOG(D_BASIC, "_make_nonblock: fcntl failed - %s\n", strerror(errno));
        return -1;
    }

    struct sockaddr_in peer;
    peer.sin_family = AF_INET;
    if (inet_aton(argv[1], &peer.sin_addr) < 0) {
        printf("ip address is in invalid form: %s\n", strerror(errno));
        return -1;
    }
    short port = atoi(argv[2]);
    peer.sin_port = htons(port);
    socklen_t addrlen = sizeof(struct sockaddr_in);
    if (DPF_connect(fd, (struct sockaddr *)&peer, addrlen) < 0) {
        if (errno != EINPROGRESS) {
            printf("<tcp_client_2> DPF_connect: failed - %s\n", strerror(errno));
            return -1;
        }
        LOG(D_FULL, "DPF_connect returned EINPROGRESS\n");
        while(1) {
            fd_set wrfds;
            FD_ZERO(&wrfds);
            FD_SET(fd, &wrfds);
            int ready = DPF_select(fd+1, NULL, &wrfds, NULL, NULL);
            if (ready < 0) {
                if (errno == EINTR) continue;
                printf("DPF_select failed: %s\n", strerror(errno));
                return -1;
            }
            LOG(D_FULL, "<tcp_client_2> DPF_select returned write ready\n");
            break;
        }
    }

    int error;
    socklen_t len = sizeof(error);
    if (DPF_getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
        printf("<tcp_client_2> DPF_getsockopt failed: %s\n", strerror(errno));
        return -1;
    }
    if (error) {
        printf("<tcp_client_2> DPF_getsockopt returned: %s\n", strerror(error));
        return -1;
    }
    printf("<tcp_client_2> connected to the server\n");

    while(1) {
        // generate a random printable character
        struct timeval tv;
        gettimeofday(&tv, NULL);
        char ch = tv.tv_usec % 24 + 40;
        char sbuf[5001], rbuf[5001];
        for (int i = 0; i < 4999; i++) {
            sbuf[i] = ch;
        }
        sbuf[5000] = '\0';
        while(1) {
            fd_set wrfds;
            FD_ZERO(&wrfds);
            FD_SET(fd, &wrfds);
            int ready = DPF_select(fd+1, NULL, &wrfds, NULL, NULL);
            if (ready < 0) {
                if (errno == EINTR) continue;
                printf("<tcp_client_2> DPF_select failed: %s\n", strerror(errno));
                return -1;
            }
            break;
        }
        int sent = DPF_send(fd, sbuf, 5000, 0);
        if (sent < 0) {
            printf("<tcp_client_2> DPF_send: %s\n", strerror(errno));
            return -1;
        }
        printf("***%d bytes sent\n", sent);
        int total = 0, rcved = 0;
        while(total < sent) {
            while(1) {
                fd_set rdfds;
                FD_ZERO(&rdfds);
                FD_SET(fd, &rdfds);
                int ready = DPF_select(fd+1, &rdfds, NULL, NULL, NULL);
                if (ready < 0) {
                    if (errno == EINTR) continue;
                    printf("DPF_select failed: %s\n", strerror(errno));
                    return -1;
                }
                break;
            }
            int rcved = DPF_recv(fd, &rbuf[total], sent - total, 0);
            if (rcved < 0) {
                printf("DPF_recv failed: %s\n", strerror(errno));
                return -1;
            } else if (rcved == 0) {
                printf("server closed\n");
                return 0;
            }
            total += rcved;
        }
        printf("%d bytes received\n", total);
        for (int i=0; i<total; i++) {
            if (sbuf[i] != rbuf[i]) {
                printf("compare failed\n");
                printf("\n\nsent: %s\n", sbuf);
                printf("\n\nrcved: %s\n", rbuf);
                return -1;
            }
        }
    }
}
