#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>

#include "VSSL.h"

// run as "echo_client_1 server_fqdn port"
int
main(int argc, char *argv[])
{
    int fd;
    struct gen_sockaddr *gaddr;
    char sbuf[10000], rbuf[10000];
    int sent, total, rcved;
    int i, outer, inner;
    unsigned long long conn_time, xtime, elapsed;
    unsigned short port;
    char ch;
    struct timeval tv;
    int pid = getpid();
    char logfile[30];
    struct timeval conn_before, conn_after, xbefore, xafter;
    FILE *exp_fp;
    int loop;

    if (argc != 4) {
        printf("usage: tcp_client_3 server_fqdn port exp_out_file\n");
        return -1;
    }
    sprintf(logfile, "clientlog.%d", pid);
    printf("logfile = %s\n", logfile);

    // open thr file to record performance data
    exp_fp = fopen(argv[3], "w");
    if (exp_fp == NULL) {
        printf("fopen failed\n");
        return -1;
    }
    fprintf(exp_fp, "=====================================================\n");
    fprintf(exp_fp, "          connection              xfer\n");
    fprintf(exp_fp, "=====================================================\n");
    printf("=====================================================\n");
    printf("          connection              xfer\n");
    printf("=====================================================\n");

    vssl_init(logfile, D_FULL, 64000, "client.pem", "root.pem", "password");

    for(outer = 0; outer < 2000; outer++) {
        conn_time = xtime = 0;
        loop = 0;
        for (inner = 0; inner < 30; inner++) {
            fd = vssl_client(TCP, BLOCK);
            gaddr = vssl_gethostbyname(argv[1]);
            if (gaddr == NULL) {
                printf("vssl_gethostbyname failed: %s\n", strerror(errno));
                return -1;
            }
            port = atoi(argv[2]);
            gaddr->addr.port = htons(port);

            if (gettimeofday(&conn_before, NULL) < 0) {
                printf("gettimeofday failed\n");
                return -1;
            }
            if (vssl_connect(fd, gaddr, BLOCK) < 0) {
                printf("vssl_connect: failed - %s\n", strerror(errno));
                break;
            }
            if (gettimeofday(&conn_after, NULL) < 0) {
                printf("gettimeofday failed\n");
                return -1;
            }
            elapsed = (conn_after.tv_sec - conn_before.tv_sec) * 1000000 +
                (conn_after.tv_usec - conn_before.tv_usec);
            conn_time += elapsed;

            sleep(1);

            // generate a random printable character
            gettimeofday(&tv, NULL);
            ch = tv.tv_usec % 24 + 40;
            for (i = 0; i < 4999; i++) {
                sbuf[i] = ch;
            }
            sbuf[5000] = '\0';
    
            if (gettimeofday(&xbefore, NULL) < 0) {
                printf("gettimeofday failed\n");
                return -1;
            }
            sent = vssl_send(fd, sbuf, 5000, 0);
            if (sent < 0) {
                printf("vssl_send: %s\n", strerror(errno));
                return -1;
            }
            //printf("***%d bytes sent: %s\n", sent, sbuf);

            total = rcved = 0;
            while(total < sent) {
                rcved = vssl_recv(fd, rbuf, sent - total, 0);
                if (rcved < 0) {
                    printf("vssl_recv failed: %s\n", strerror(errno));
                    return -1;
                } else if (rcved == 0) {
                    printf("server closed\n");
                    return 0;
                }
                total += rcved;
            }
            //printf("%d bytes received: %s\n", total, rbuf);
            if (gettimeofday(&xafter, NULL) < 0) {
                printf("gettimeofday failed\n");
                return -1;
            }
            elapsed = (xafter.tv_sec - xbefore.tv_sec) * 1000000 +
                (xafter.tv_usec - xbefore.tv_usec);
            xtime += elapsed;
    
            for (i=0; i<total; i++) {
                if (sbuf[i] != rbuf[i]) {
                    printf("compare failed\n");
                    return -1;
                }
            }
    
            vssl_close(fd);
            loop++;

            sleep(1);
        } // inner
        if (loop != 0) {
            conn_time = conn_time / loop;
            xtime = xtime / loop;
            fprintf(exp_fp, "%lld      %lld\n", conn_time, xtime);
            fflush(exp_fp);
            printf("%lld      %lld\n", conn_time, xtime);
        }
    } // outer
    fprintf(exp_fp, "=====================================================\n");
    printf("=====================================================\n");
    fclose(exp_fp);

    return 0;
}
