#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "DPF.h"
#include "internet.h"

// run as "tcp_server_1 port"
int
main(int argc, char *argv[])
{
    if (argc != 2) {
        printf("usage: tcp_server_1 port\n");
        return -1;
    }

    int fd = DPF_socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        printf("DPF_socket failed: %s\n", strerror(errno));
        return -1;
    }

    short port = atoi(argv[1]);
    uint16_t nport = htons(port);
    struct sockaddr_in myaddr;
    myaddr.sin_family = AF_INET;
    myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    myaddr.sin_port = nport;
    if (DPF_bind(fd, (struct sockaddr *)&myaddr, sizeof(myaddr)) < 0) {
        printf("DPF_bind failed: %s\n", strerror(errno));
        return -1;
    }
    if (DPF_listen(fd, 5) < 0) {
        printf("DPF_listen: failed - %s\n", strerror(errno));
        return -1;
    }
    socklen_t addrlen = sizeof(struct sockaddr_in);
    if (DPF_getsockname(fd, (struct sockaddr *)&myaddr, &addrlen) < 0) {
        printf("DPF_getsockname failed - %s\n", strerror(errno));
        return -1;
    }
    printf("listening on %s\n", _sin2str(&myaddr));

    struct sockaddr_in peer;
    addrlen = sizeof(struct sockaddr_in);
    int newfd = DPF_accept(fd, (struct sockaddr *)&peer, &addrlen);
    if (newfd < 0) {
        printf("DPF_accept: failed - %s\n", strerror(errno));
        return -1;
    }
    printf("connection accepted from %s\n", _sin2str(&peer));
    printf("closing the listening socket\n");
    DPF_close(fd);

    while(1) {
        char buf[1000];
        int bytes = DPF_recv(newfd, buf, 1000, 0);
        if (bytes < 0) {
            printf("DPF_recv failed: %s\n", strerror(errno));
            return -1;
        } else if (bytes == 0) {
            DPF_close(newfd);
            printf("done\n");
            return 0;
        }
        printf("\t%d bytes received\n", bytes);
        int total = 0, sent = 0;
        while(total < bytes) {
            sent = DPF_send(newfd, &buf[total], bytes - total, 0);
            if (sent < 0) {
                printf("DPF_send failed: %s\n", strerror(errno));
                return -1;
            }
            total += sent;
        }
        printf("\t%d bytes sent\n", total);
    }
}
