#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "internet.h"
#include "DPF.h"

static int
echo(int fd)
{
    int bytes;
    int total, sent;
    char buf[10000];

    bytes = DPF_recv(fd, buf, 10000, 0);
    if (bytes < 0) {
        printf("DPF_recv failed: %s\n", strerror(errno));
        return -1;
    } else if (bytes == 0) {
        printf("client closed\n");
        return 0;
    }
    printf("\t%d bytes received\n", bytes);
    total = sent = 0;
    while(total < bytes) {
        sent = DPF_send(fd, &buf[total], bytes - total, 0);
        if (sent < 0) {
            printf("DPF_send failed: %s\n", strerror(errno));
            return -1;
        }
        total += sent;
    }
    printf("\techoed\n");
    return total;
}

// run as "tcp_server_1 port"
int
main(int argc, char *argv[])
{
    int fd, newfd, maxfd;
    short port;
    uint16_t nport;
    struct sockaddr_in myaddr;
    fd_set rdfds, wrfds;
    int error;
    int ready;
    char fds[1000];
    int i;

    for (i=0; i<1000; i++) {
        fds[i] = 0;
    }

    if (argc != 2) {
        printf("usage: tcp_server_1 port\n");
        return -1;
    }

    fd = DPF_socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        printf("DPF_socket failed: %s\n", strerror(errno));
        return -1;
    }
    printf("listen sock: fd = %d\n", fd);

    myaddr.sin_family = AF_INET;
    myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    port = atoi(argv[1]);
    myaddr.sin_port = htons(port);
    if (DPF_bind(fd, (struct sockaddr *)&myaddr, sizeof(myaddr)) < 0) {
        printf("DPF_bind failed: %s\n", strerror(errno));
        return -1;
    }
    if (DPF_listen(fd, 5) < 0) {
        printf("DPF_listen: failed - %s\n", strerror(errno));
        return -1;
    }
    socklen_t addrlen = sizeof(struct sockaddr_in);
    if (DPF_getsockname(fd, (struct sockaddr *)&myaddr, &addrlen) < 0) {
        printf("DPF_getsockname failed - %s\n", strerror(errno));
        return -1;
    }
    printf("listening on %s\n", _sin2str(&myaddr));
    fds[fd] = 1;

    while(1) {
        // select on listen sock and accepted socks
        FD_ZERO(&rdfds);
        maxfd = 0;
        for (i=0; i<1000; i++) {
            if (fds[i]) {
                FD_SET(i, &rdfds);
                maxfd = i + 1;
            }
        }
        ready = DPF_select(maxfd, &rdfds, NULL, NULL, NULL);
        if (ready < 0) {
            if (errno == EINTR) continue;
            printf("DPF_select failed: %s\n", strerror(errno));
            return -1;
        }
        for (i=0; i<maxfd; i++) {
            if (i == fd) continue;
            if (FD_ISSET(i, &rdfds)) {
                if (echo(i) <= 0) {
                    DPF_close(i);
                    fds[i] = 0;
                }
            }
        }
        if (FD_ISSET(fd, &rdfds)) {
            struct sockaddr_in peer;
            addrlen = sizeof(peer);
            newfd = DPF_accept(fd, (struct sockaddr *)&peer, &addrlen);
            if (newfd < 0) {
                printf("DPF_accept: failed - %s\n", strerror(errno));
                return -1;
            }
            printf("connection accepted from %s\n", _sin2str(&peer));
            fds[newfd] = 1;
        }
    }
}
