#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "internet.h"
#include "DPF.h"

// run as "udp_server_2 port"
int
main(int argc, char *argv[])
{
    int fd;
    short port;
    uint16_t nport;
    char buf[1000];
    int bytes, total, sent;
    fd_set rdfds;
    int ready;

    if (argc != 2) {
        printf("usage: udp_server_2 port\n");
        return -1;
    }

    fd = DPF_socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0) {
        printf("DPF_socket failed: %s\n", strerror(errno));
        return -1;
    }
    printf("listen sock: fd = %d\n", fd);

    struct sockaddr_in myaddr;
    myaddr.sin_family = AF_INET;
    myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    port = atoi(argv[1]);
    myaddr.sin_port = htons(port);
    if (DPF_bind(fd, (struct sockaddr *)&myaddr, sizeof(myaddr)) < 0) {
        printf("DPF_bind failed: %s\n", strerror(errno));
        return -1;
    }

    socklen_t addrlen = sizeof(struct sockaddr_in);
    if (DPF_getsockname(fd, (struct sockaddr *)&myaddr, &addrlen) < 0) {
        printf("DPF_getsockname failed - %s\n", strerror(errno));
        return -1;
    }
    printf("listening on %s\n", _sin2str(&myaddr));

    while(1) {
        FD_ZERO(&rdfds);
        FD_SET(fd, &rdfds);
        ready = DPF_select(fd+1, &rdfds, NULL, NULL, NULL);
        if (ready < 0) {
            if (errno == EINTR) { continue; }
            printf("select failed: %s\n", strerror(errno));
            return -1;
        }
        struct sockaddr_in from;
        socklen_t fromlen = sizeof(from);
        bytes = DPF_recvfrom(fd, buf, 1000, 0,
                (struct sockaddr *)&from, &fromlen);
        if (bytes < 0) {
            printf("DPF_recvfrom failed: %s\n", strerror(errno));
            return -1;
        }
        printf("\t%d bytes received from: %s\n", bytes, _sin2str(&from));
        printf("\tcontent: %s\n", buf);
        sent = DPF_sendto(fd, buf, bytes, 0,
            (struct sockaddr *)&from, fromlen);
        if (sent < 0) {
            printf("DPF_sendto failed: %s\n", strerror(errno));
            return -1;
        }
        printf("\treplied: %s\n", buf);
    }
}
