/****************************************************************************
 * Please read "../../LICENSE.TXT" for copyright
 ****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "log.h"

// This file is far from complete
// The main goal at this moment is that Generic socket can log using Condor's
// dprint and this file just provide a place that Condor's dprintf can be plugged in

int Generic_log_initialized = 0;
int Generic_log_enabled = 1;
char Generic_logfile[50];
int Generic_debug_level = D_BASIC;
int Generic_maxlog = 64000;
char *Generic_log_srcfile;
int Generic_log_lineno;
// the pointer to the actual log function
static void default_log_va(int level, char *fmt, va_list ap);
void (*Generic_log_va)(int flag, char *fmt, va_list args) = default_log_va;

void
Generic_set_log_va(void(*app_log_va)(int level, char *fmt, va_list args))
{
	Generic_log_va = app_log_va;
}

void
Generic_logInit(char *path, int debug_level, int maxsize)
{
	char *namestr, *levelstr, *sizestr;
	char msg[1000];
	int fd;

	Generic_log_initialized = 1;

	// Compose a logfile name and open the logfile
	if (path) {
		namestr = path;
		strcpy(Generic_logfile, path);
	} else {
		namestr = getenv("SS_LOG_FILE");
		if (namestr) {
			strcpy(Generic_logfile, namestr);
		}
	}
	if (namestr) {
		fd = open(Generic_logfile, O_RDWR | O_CREAT | O_APPEND, S_IRWXU);
		if (fd < 0){
			Generic_logfile[0] = '\0';
			return;
		}
	} else {
		return;
	}

	// Set log level
	if (debug_level >= 0) {
		Generic_debug_level = debug_level;
	} else if ((levelstr = getenv("SS_LOG_LEVEL")) != NULL) {
		if (!strcasecmp(levelstr, "fulldebug")) {
			Generic_debug_level = D_FULL;
		}
	}

	// Set the maximum size of the logfile
	if (maxsize > 0) {
		Generic_maxlog = maxsize;
	} else {
		sizestr = getenv("SS_LOG_MAXSIZE");
		if (sizestr) {
			Generic_maxlog = atol(sizestr);
		}
	}

	sprintf(msg, "********************************************************\n");
	sprintf(&msg[strlen(msg)], "\tNew log file started. Max size = %d\n", Generic_maxlog);
	switch(Generic_debug_level) {
		case D_FULL: sprintf(&msg[strlen(msg)], "\t\tLog level: D_FULL\n"); break;
		default:
			sprintf(&msg[strlen(msg)], "\t\tLog level: D_BASIC\n"); break;
	}
	sprintf(&msg[strlen(msg)], "********************************************************\n");
	if (write(fd, msg, strlen(msg)) < 0) {
		// if the very first log failed, we disable log
		Generic_logfile[0] = '\0';
	}
	close(fd);
	return;
}

static void
default_log_va(int level, char *fmt, va_list ap)
{
	char str[4000], oldname[50];
	struct tm *t, *localtime();
    time_t clock;
	struct stat statbuf;
	int fd;

	if (!Generic_log_initialized) Generic_logInit(NULL, -1, -1);

	if (Generic_logfile[0] == '\0') return;
	if (level > Generic_debug_level) return;

	// Check the status of logfile
	if (stat(Generic_logfile, &statbuf) < 0) {
		if (errno != ENOENT) {
			// disable log
			Generic_logfile[0] = '\0';
			return;
		}
	}

	// If the logfile is too big, rename the file to file.old
	if (statbuf.st_size > Generic_maxlog) {
		sprintf(oldname, "%s.old", Generic_logfile);
		(void) rename(Generic_logfile, oldname);
	}

	// Make log message
	(void)time(  (time_t *)&clock );
    t = localtime( (time_t *)&clock );
	sprintf(str, "<%d/%d %02d:%02d:%02d> [%s:%d] (%d)",
			t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec,
			Generic_log_srcfile, Generic_log_lineno, getpid());
	vsprintf(&str[strlen(str)], fmt, ap);

	// Now open the file and write the message to it
	fd = open(Generic_logfile, O_RDWR | O_CREAT | O_APPEND, S_IRWXU);
	if (fd < 0) {
		return;
	}
	(void) write(fd, str, strlen(str));
	(void) close(fd);
	return;
}

void
Generic_stop_logging()
{
	Generic_log(D_BASIC, "For process [pid=%d], logging has been stopped\n", getpid());
    Generic_log_enabled = 0;
}

void
Generic_resume_logging()
{
	Generic_log(D_BASIC, "For process [pid=%d], logging has been resumed\n", getpid());
    Generic_log_enabled = 1;
}

void
Generic_log(int level, char *fmt, ...)
{
    if (!Generic_log_enabled) { return; }

	va_list ap;
	va_start(ap, fmt);
	(*Generic_log_va)(level, fmt, ap);
	va_end(ap);
}

void
Generic_error(char *fmt, ...)
{
	va_list ap;
	char str[4000];

	va_start(ap, fmt);
	vsprintf(str, fmt, ap);
	va_end(ap);
    if (errno != 0) {
		sprintf(&str[strlen(str)], "[%s]", strerror(errno));
    }
	Generic_log(D_BASIC, "\n[%d] ERROR %s:%d: \"%s\"\n", getpid(),
            Generic_log_srcfile, Generic_log_lineno, str);
	return;
}

void
Generic_abort(char *fmt, ...)
{
    va_list ap;
	char str[4000];

	va_start(ap, fmt);
	vsprintf(str, fmt, ap);
	va_end(ap);
    if (errno != 0) {
		sprintf(&str[strlen(str)], "[%s]", strerror(errno));
    }
	Generic_log(D_BASIC, "\n[%d] ERROR %s:%d: \"%s\"\n", getpid(),
            Generic_log_srcfile, Generic_log_lineno, str);
	exit(-1);
}

void
Generic_job_log(char *msg)
{
    char str[4000];
    struct tm *t, *localtime();
    time_t clock;
    int fd;

    // Make log message
    (void)time(  (time_t *)&clock );
    t = localtime( (time_t *)&clock );
    sprintf(str, "<%d/%d %02d:%02d:%02d> [%s:%d] (%d)",
            t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec,
            Generic_log_srcfile, Generic_log_lineno, getpid());
    sprintf(&str[strlen(str)], msg);

    // Now open the file and write the message to it
    fd = open("/tmp/Genericlog", O_RDWR | O_CREAT | O_APPEND, S_IRWXU|S_IRWXG|S_IRWXO);
    if (fd < 0) {
        return;
    }
    (void) write(fd, str, strlen(str));
    (void) close(fd);
    return;
}
