#ifndef _Generic_LOG_H
#define _Generic_LOG_H

#include <stdarg.h>
#include <sys/time.h>

#ifdef __cplusplus
extern "C" {
#endif

extern char *Generic_log_srcfile;
extern int Generic_log_lineno;

#define D_BASIC (1 << 0)
#define D_FULL (1 << 13)

#define LOG \
	Generic_log_srcfile = __FILE__;\
	Generic_log_lineno = __LINE__;\
	Generic_log
#define ERROR \
	Generic_log_srcfile = __FILE__;\
	Generic_log_lineno = __LINE__;\
	Generic_error
#define ABORT \
    Generic_log_srcfile = __FILE__;\
    Generic_log_lineno = __LINE__;\
    Generic_abort

#define JOB_LOG \
    Generic_log_srcfile = __FILE__;\
    Generic_log_lineno = __LINE__;\
	Generic_job_log

// Set log file and log level to 'path' and 'level', respectively
// Before this routine is called, every log will be printed to stdout
void Generic_logInit(char *path, int level, int maxsize);

// With this function, application let Generic use its own function to log.
// The only application that is expected to use is Condor at this moment.
// Since Generic_log, Generic_error, and Generic_abort will eventuall call the function
// passed by the application by calling this, it should still call LOG, ERROR,
// and ABORT
void Generic_set_log_va(void(*app_log_va)(int level, char *fmt, va_list args));

void Generic_stop_logging();
void Generic_resume_logging();

// Print debug message of debug level 'level'. If 'level' is higher
// than 'Generic_debug_level', the msg will not be logged
void Generic_log(int level, char *fmt, ...);

// Print error message to the log file and exit
// The error message passed will be prefixed with pid, file, and line where error occured
// and also postfixed with strerror(errno)
void Generic_abort(char *fmt, ...);

// Same as Generic_abort without exiting
void Generic_error(char *fmt, ...);

void Generic_job_log(char *msg);

#ifdef __cplusplus
}
#endif

#endif
