/****************************************************************************
 * Please read "../../LICENSE.TXT" for copyright
 ****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "log.h"

// This file is far from complete
// The main goal at this moment is that GCB can log using Condor's dprint and
// this file just provide a place that Condor's dprintf can be plugged in

int _ss_log_initialized = 0;
char _ss_logfile[50];
int _ss_debug_level = D_BASIC;
int _ss_maxlog = 64000;
char *_ss_log_srcfile;
int _ss_log_lineno;
// the pointer to the actual log function
static void default_log_va(int level, char *fmt, va_list ap);
void (*_ss_log_va)(int flag, char *fmt, va_list args) = default_log_va;

void
_ss_set_log_va(void(*app_log_va)(int level, char *fmt, va_list args))
{
    _ss_log_va = app_log_va;
}

void
_ss_logInit(char *path, int debug_level, int maxsize)
{
    char *namestr, *levelstr, *sizestr;
    char msg[1000];
    int fd;

    _ss_log_initialized = 1;

    // Compose a logfile name and open the logfile
    if (path) {
        namestr = path;
        strcpy(_ss_logfile, path);
    } else {
        namestr = getenv("SS_LOG_FILE");
        if (namestr) {
            strcpy(_ss_logfile, namestr);
        }
    }
    if (namestr) {
        fd = open(_ss_logfile, O_RDWR | O_CREAT | O_APPEND, S_IRWXU);
        if (fd < 0){
            _ss_logfile[0] = '\0';
            return;
        }
    } else {
        return;
    }

    // Set log level
    if (debug_level >= 0) {
        _ss_debug_level = debug_level;
    } else if ((levelstr = getenv("SS_LOG_LEVEL")) != NULL) {
        if (!strcasecmp(levelstr, "fulldebug")) {
            _ss_debug_level = D_FULL;
        }
    }

    // Set the maximum size of the logfile
    if (maxsize > 0) {
        _ss_maxlog = maxsize;
    } else {
        sizestr = getenv("SS_LOG_MAXSIZE");
        if (sizestr) {
            _ss_maxlog = atol(sizestr);
        }
    }

    sprintf(msg, "********************************************************\n");
    sprintf(&msg[strlen(msg)], "\tNew log file started. Max size = %d\n", _ss_maxlog);
    switch(_ss_debug_level) {
        case D_FULL: sprintf(&msg[strlen(msg)], "\t\tLog level: D_FULL\n"); break;
        default:
            sprintf(&msg[strlen(msg)], "\t\tLog level: D_BASIC\n"); break;
    }
    sprintf(&msg[strlen(msg)], "********************************************************\n");
    if (write(fd, msg, strlen(msg)) < 0) {
        // if the very first log failed, we disable log
        _ss_logfile[0] = '\0';
    }
    close(fd);
    return;
}

static void
default_log_va(int level, char *fmt, va_list ap)
{
    char str[4000], oldname[50];
    struct tm *t, *localtime();
    time_t clock;
    struct stat statbuf;
    int fd;

    if (!_ss_log_initialized) _ss_logInit(NULL, -1, -1);

    if (_ss_logfile[0] == '\0') return;
    if (level > _ss_debug_level) return;

    // Check the status of logfile
    if (stat(_ss_logfile, &statbuf) < 0) {
        if (errno != ENOENT) {
            // disable log
            _ss_logfile[0] = '\0';
            return;
        }
    }

    // If the logfile is too big, rename the file to file.old
    if (statbuf.st_size > _ss_maxlog) {
        sprintf(oldname, "%s.old", _ss_logfile);
        (void) rename(_ss_logfile, oldname);
    }

    // Make log message
    (void)time(  (time_t *)&clock );
    t = localtime( (time_t *)&clock );
    sprintf(str, "<%d/%d %02d:%02d:%02d> [%s:%d] ",
            t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec,
            _ss_log_srcfile, _ss_log_lineno);
    vsprintf(&str[strlen(str)], fmt, ap);

    // Now open the file and write the message to it
    fd = open(_ss_logfile, O_RDWR | O_CREAT | O_APPEND, S_IRWXU);
    if (fd < 0) {
        return;
    }
    (void) write(fd, str, strlen(str));
    (void) close(fd);
    return;
}

void
_ss_log(int level, char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    (*_ss_log_va)(level, fmt, ap);
    va_end(ap);
}

void
_ss_error(char *fmt, ...)
{
    va_list ap;
    char str[4000];

    va_start(ap, fmt);
    vsprintf(str, fmt, ap);
    va_end(ap);
    if (errno != 0) {
        sprintf(&str[strlen(str)], "[%s]", strerror(errno));
    }
    _ss_log(D_BASIC, "\n[%d] ERROR %s:%d: \"%s\"\n", getpid(), _ss_log_srcfile, _ss_log_lineno, str);
    return;
}

void
_ss_abort(char *fmt, ...)
{
    va_list ap;
    char str[4000];

    va_start(ap, fmt);
    vsprintf(str, fmt, ap);
    va_end(ap);
    if (errno != 0) {
        sprintf(&str[strlen(str)], "[%s]", strerror(errno));
    }
    _ss_log(D_BASIC, "\n[%d] ERROR %s:%d: \"%s\"\n", getpid(), _ss_log_srcfile, _ss_log_lineno, str);
    exit(-1);
}
