#ifndef __SS_LOG_H
#define __SS_LOG_H

#include <stdarg.h>
#include <sys/time.h>

#ifdef __cplusplus
extern "C" {
#endif

extern char *_ss_log_srcfile;
extern int _ss_log_lineno;

#define D_BASIC (1 << 0)
#define D_FULL (1 << 13)

#define LOG \
    _ss_log_srcfile = __FILE__;\
    _ss_log_lineno = __LINE__;\
    _ss_log
#define ERROR \
    _ss_log_srcfile = __FILE__;\
    _ss_log_lineno = __LINE__;\
    _ss_error
#define ABORT \
    _ss_log_srcfile = __FILE__;\
    _ss_log_lineno = __LINE__;\
    _ss_abort

// Set log file and log level to 'path' and 'level', respectively
// Before this routine is called, every log will be printed to stdout
void _ss_logInit(char *path, int level, int maxsize);

// With this function, application let GCB use its own function to log.
// The only application that is expected to use is Condor at this moment.
// Since _ss_log, _ss_error, and _ss_abort will eventuall call the function
// passed by the application by calling this, it should still call LOG, ERROR,
// and ABORT
void _ss_set_log_va(void(*app_log_va)(int level, char *fmt, va_list args));

// Print debug message of debug level 'level'. If 'level' is higher
// than '_ss_debug_level', the msg will not be logged
void _ss_log(int level, char *fmt, ...);

// Print error message to the log file and exit
// The error message passed will be prefixed with pid, file, and line where error occured
// and also postfixed with strerror(errno)
void _ss_abort(char *fmt, ...);

// Same as _ss_abort without exiting
void _ss_error(char *fmt, ...);

#ifdef __cplusplus
}
#endif

#endif
