#ifndef _nb_wbuf
#define _nb_wbuf

// This class provides a buffering mechanism for nonblocking sending
class NB_wbuf {
    private:
    public:
        const static unsigned short MAX = 1000;
        int fd;
        unsigned char data[MAX];
        int size; // the size of the data buffered in 'data'
        int bytes;// the amount of data sent so far

        NB_wbuf(int fd);
        bool empty();
        // Push the content of data[bytes ~ size] to the network. Caller must
        // make sure that the buffer is not empty by calling empty().
        // @ret:
        //    - true if the buffer is totally flushed
        //    - false otherwise w/ errno set to:
        //      + no error occurred, but partially flushed: EAGAIN
        //      + send fails: errno returned by send(including EAGAIN)
        bool flush();
        // Push the record to the network as much as possible by a single send,
        // and then put leftover into 'data'
        // This routine overwrites previous msg, so, caller must make sure that
        // the previous msg was fully flushed into the TCP buffer by calling
        // can_send()
        // @args:
        //   - record: the char array holding the record to send.
        //   - len: the size of the record. Should not be greater than the
        //          size of 'data[]'
        // @ret:
        //    - true if the record is totally sent
        //    - false otherwise w/ errno set to:
        //      + no error occurred, but partially sent and remaining part buffered: EAGAIN
        //      + send fails: errno returned by send(including EAGAIN)
        //      + len > bufsize: EINVAL
        bool send(char *record, int len);
        inline void set_fd(int newfd) { fd = newfd; }
        void tostr(char *str);
        int stripe(char *buf);
        int unstripe(char *buf);
};

#endif
